import troposphere.kms as kms

from shared import *


class KMSKey(kms.Key):
    def setup(self):
        self.KeyPolicy = {
            'Version': '2012-10-17',
            'Id': 'key-default-1',
            'Statement': [
                {
                    'Action': 'kms:*',
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': Sub('arn:aws:iam::${AWS::AccountId}:root')
                    },
                    'Resource': '*',
                    'Sid': 'Enable IAM User Permissions'
                },
            ]
        }


class KMSAliasParameterStore(kms.Alias):
    def setup(self):
        self.AliasName = Sub('alias/parameter_store_key')
        self.TargetKeyId = Ref('KMSKeyParameterStore')

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class KMS_Keys(object):
    def __init__(self, key):
        # Resources
        R_KeyParameterStore = KMSKey('KMSKeyParameterStore')
        R_KeyParameterStore.setup()
        
        R_AliasParameterStore = KMSAliasParameterStore('KMSAliasParameterStore')
        R_AliasParameterStore.setup()

        cfg.Resources.extend([
            R_KeyParameterStore,
            R_AliasParameterStore,
        ])

        # Outputs
        O_ParameterStore = Output('KeyParameterStore')
        O_ParameterStore.Value = Sub('${KMSKeyParameterStore.Arn}')
        O_ParameterStore.Export = Export('KeyParameterStore')

        cfg.Outputs.extend([
            O_ParameterStore,
        ])

# Need to stay as last lines
import_modules(globals())



import sys
import os
import copy
from collections import OrderedDict, Mapping
from pprint import pprint, pformat

import cfg


def get_final_resources(resources):
    resources['final'] = {}

    for r in ['cmm', stacktype, envrole]:
        if r in resources:
            resources['final'].update(resources[r])

    return resources['final']


def mappings_create_entry(mappings, keyenv, keyregion, keyname, keyvalue):
    # create mapping entry
    mappings[keyenv][keyregion].update({keyname: keyvalue})
    # and if present delete the common one, but first record its value
    if keyenv != 'cmm' and keyname in mappings['cmm']['cmm']:
        value = mappings['cmm']['cmm'][keyname]
        del mappings['cmm']['cmm'][keyname]
        # for all real env region that do not already have a key entry i need to create a mapping entry equals to the common one
        # for the non already processed one it could be rewritten later.
        for env in RP_base:
            for region in RP_base[env]:
                if env != 'cmm' and keyname not in mappings[env][region]:
                    mappings[env][region].update({keyname: value})


def get_mapping_env_region(MP, RP, e, r, p):
    mappings = MP

    if isinstance(RP, dict):
        for i, v in RP.iteritems():
            if not e and not r:
                get_mapping_env_region(mappings, v, i, None, None)
            elif not r:
                get_mapping_env_region(mappings, v, e, i, None)
            elif not p:
                get_mapping_env_region(mappings, v, e, r, i)
            else:
                get_mapping_env_region(mappings, v, e, r, str(p) + str(i))
    else:
        if p in mappings['cmm']['cmm'] and RP == mappings['cmm']['cmm'][p]:
            pass
        else:
            mappings_create_entry(mappings, e, r, p, RP)

    return mappings


def get_mappings(RP):
    global mappedvalue

    mappings = {}
    mappings['cmm'] = copy.deepcopy(RP_base)
    mappings['ec2'] = {}
    mappings['resources-env'] = {}

    mappings['cmm'] = get_mapping_env_region(copy.deepcopy(RP_base), RP, None, None, None)

    mappedvalue = mappings['cmm']['cmm']['cmm']
    cfg.mappedvalue = mappedvalue
    # pprint(mappedvalue)

    # delete empy mappings, CloudFormation do not like them!
    for env in RP_base:
        for region in RP_base[env]:
            if len(mappings['cmm'][env][region]) == 0:
                del mappings['cmm'][env][region]
        if len(mappings['cmm'][env]) == 0:
            del mappings['cmm'][env]

    # and remove no more needed cmm one
    if 'cmm' in mappings['cmm']:
        del mappings['cmm']['cmm']

#   #START##MAP INSTANCE TYPE EPHEMERAL####
    map_eph = {
        'InstaceEphemeral0': ['m3.medium', 'm3.large', 'm3.xlarge', 'm3.2xlarge', 'c3.large',
                              'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge', 'g2.2xlarge', 'r3.large', 'r3.xlarge',
                              'r3.2xlarge', 'r3.4xlarge', 'i2.xlarge', 'i2.2xlarge', 'i2.4xlarge',
                              'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge'],
        'InstaceEphemeral1': ['m3.xlarge', 'm3.2xlarge', 'c3.large', 'c3.xlarge', 'c3.2xlarge', 'c3.4xlarge',
                              'i2.2xlarge', 'i2.4xlarge', 'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge'],
        'InstaceEphemeral2': ['i2.4xlarge', 'd2.xlarge', 'd2.2xlarge', 'd2.4xlarge']
    }

    mappings['ec2']['InstanceTypes'] = {}
    instance_list = list(
        set(map_eph['InstaceEphemeral0'] + map_eph['InstaceEphemeral1'] + map_eph['InstaceEphemeral2'])
    )
    for i in instance_list + ['default',
                              't2.nano', 't2.micro', 't2.small', 't2.medium', 't2.large', 't2.xlarge', 't2.2xlarge',
                              't3.nano', 't3.micro', 't3.small', 't3.medium', 't3.large', 't3.xlarge', 't3.2xlarge',
                              'm4.large', 'm4.xlarge', 'm4.2xlarge', 'm4.4xlarge',
                              'm5.large', 'm5.xlarge', 'm5.2xlarge', 'm5.4xlarge',
                              'c4.large', 'c4.xlarge', 'c4.2xlarge', 'c4.4xlarge',
                              'c5.large', 'c5.xlarge', 'c5.2xlarge', 'c5.4xlarge',
                              'r4.large', 'r4.xlarge', 'r4.2xlarge', 'r4.4xlarge',
                              'r5.large', 'r5.xlarge', 'r5.2xlarge', 'r5.4xlarge',
                              'g3s.xlarge']:
        mappings['ec2']['InstanceTypes'].update({
            i: {
                'InstaceEphemeral0': '1' if i in map_eph['InstaceEphemeral0'] else '0',
                'InstaceEphemeral1': '1' if i in map_eph['InstaceEphemeral1'] else '0',
                'InstaceEphemeral2': '1' if i in map_eph['InstaceEphemeral2'] else '0',
            }
        })

#   #END##MAP INSTANCE TYPE EPHEMERAL####

#  #START##MAP AVABILITY ZONES####
    mappings['resources-env']['AvabilityZones'] = {
        'eu-west-1': {
            'Zone0': 'True',
            'Zone1': 'True',
            'Zone2': 'True'
        },
        'eu-central-1': {
            'Zone0': 'True',
            'Zone1': 'True',
            'Zone2': 'False'
        },
    }
#  #END##MAP AVABILITY ZONES####

    return get_final_resources(mappings)


global envrole
global stacktype
global RP_base
global mappings

envrole = cfg.envrole
stacktype = cfg.stacktype
RP_base = cfg.RP_base

mappings = get_mappings(cfg.RP)

if mappings is not None:
    for m, v in mappings.iteritems():
        cfg.template.add_mapping(m, v)



import troposphere.sqs as sqs

from shared import *


class SQSQueue(sqs.Queue):
    def setup(self):
        self.MessageRetentionPeriod = 360


class SQSQueuePolicy(sqs.QueuePolicy):
    def setup(self, key):
        self.Queues = [
            Sub(
                'https://sqs.${AWS::Region}.amazonaws.com/${AWS::AccountId}/${QueueId}',
                **{'QueueId': Select('5', Split(':', eval(key['Endpoint'])))}
            )
        ]
        self.PolicyDocument = {
            'Version': '2012-10-17',
            'Statement': [
                {   
                    'Action': [
                        'SQS:SendMessage'
                    ],
                    'Condition': {
                        'ArnEquals': {
                            'aws:SourceArn': eval(key['TopicArn'])
                        }
                    },
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': '*'
                    },
                    'Resource': '*'
                }
            ]
        }

##
##

class SQS_Queues(object):
    def __init__(self, key):
        for n, v in RP_cmm[key].iteritems():
            resname = key + n
            # resources
            r_Queue = SQSQueue(resname)
            r_Queue.setup()

            cfg.Resources.extend([
                r_Queue,
            ])

# Need to stay as last lines
import_modules(globals())



import troposphere.ecs as ecs
import troposphere.ecr as ecr

from shared import *


class ECSService(ecs.Service):
    def setup(self, scheme):
        self.Cluster = get_exported_value('Cluster', 'ClusterStack')
        # DAEMON MODE DO NOT SUPPORT DESIRED OR PLACEMENT STRATEGIES, IS SIMPLY ONE TASK FOR CONTAINER INSTANCE
        if 'SchedulingStrategy' not in RP_cmm or RP_cmm['SchedulingStrategy'] == 'REPLICA':
            self.DesiredCount = get_final_value('CapacityDesired')
            self.DeploymentConfiguration = ecs.DeploymentConfiguration(
                MaximumPercent = get_final_value('DeploymentConfigurationMaximumPercent'),
                MinimumHealthyPercent = get_final_value('DeploymentConfigurationMinimumHealthyPercent'),
            )
            self.PlacementStrategies = If(
                'LaunchTypeFarGate',
                Ref('AWS::NoValue'),
                [
                    #ecs.PlacementStrategy(
                    #    Type='spread',
                    #    Field='attribute:ecs.availability-zone'
                    #),
                    ecs.PlacementStrategy(
                        Type='spread',
                        Field='instanceId'
                    )
                ]
            )
        elif RP_cmm['SchedulingStrategy'] == 'DAEMON':
            self.SchedulingStrategy = 'DAEMON'

        self.LaunchType = get_final_value('LaunchType')
        if 'HealthCheckGracePeriodSeconds' in RP_cmm:
            self.HealthCheckGracePeriodSeconds = get_final_value('HealthCheckGracePeriodSeconds')
        if scheme:
            self.LoadBalancers = [
                ecs.LoadBalancer(
                    ContainerName=Ref('EnvRole'),
                    ContainerPort=get_final_value('ContainerDefinitions1ContainerPort'),
                    TargetGroupArn=Ref('TargetGroup' + scheme)
                )
            ]
        #self.PlatformVersion = 'LATEST'
        if scheme:
            self.Role = If(
                'NetworkModeAwsVpc',
                Ref('AWS::NoValue'),
                get_exported_value('RoleECSService')
            )
        self.TaskDefinition = Ref('TaskDefinition')


class ECSTaskDefinition(ecs.TaskDefinition):
    def setup(self):
        self.ExecutionRoleArn = If(
            'LaunchTypeFarGate',
            get_exported_value('RoleECSTaskExecution'),
            Ref('AWS::NoValue')
        )
        self.TaskRoleArn = Ref('RoleTask')
        if 'Cpu' in RP_cmm:
            self.Cpu = If(
                'CpuTask',
                get_final_value('Cpu'),
                Ref('AWS::NoValue')
            )
        if 'Memory' in RP_cmm:
            self.Memory = If(
                'LaunchTypeFarGate',
                get_final_value('Memory'),
                Ref('AWS::NoValue')
            )
        self.NetworkMode = If(
            'NetworkModeAwsVpc',
            'awsvpc',
            Ref('AWS::NoValue')
        )
        self.RequiresCompatibilities = If(
            'LaunchTypeFarGate',
            ['EC2', 'FARGATE'],
            ['EC2']
        )
        

class ECRRepositories(ecr.Repository):
    def setup(self):
        self.RepositoryName = get_final_value(self.title)
        self.RepositoryPolicyText = {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': [
                        'ecr:GetDownloadUrlForLayer',
                        'ecr:BatchGetImage',
                        'ecr:BatchCheckLayerAvailability',
                        'ecr:PutImage',
                        'ecr:InitiateLayerUpload',
                        'ecr:UploadLayerPart',
                        'ecr:CompleteLayerUpload',
                        'ecr:ListImages',
                        'ecr:DescribeRepositories',
                        'ecr:DescribeImages'
                    ],
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': [
                            Sub('arn:aws:iam::${AWS::AccountId}:root')
                        ]
                    },
                    'Sid': 'AllowPushPull'
                },
            ]
        }

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

class ECSEnvironment(ecs.Environment):
    def setup(self, key):
        name = self.title
        valuename = name + 'Value'
        self.Name = get_final_value(name + 'Name')
        self.Value = get_final_value(valuename)


class ECSMountPoint(ecs.MountPoint):
    def __init__(self, title, key, **kwargs):
        super(ECSMountPoint, self).__init__(title, **kwargs)
        self.ReadOnly = False
        self.SourceVolume = self.title
        self.ContainerPath = key['ContainerPath']


class ECSContainerDefinition(ecs.ContainerDefinition):
    def setup(self, key, index):
        name = self.title  # Ex. ContainerDefinitions1
        auto_get_props(self, key)

        self.Essential = True

        if len(RP_cmm['ContainerDefinitions']) == 1:
            self.Cpu = If(
                'CpuTask',
                get_final_value('Cpu'),
                get_final_value(name + 'Cpu')
            )
            self.Memory = If(
                'LaunchTypeFarGate',
                get_final_value('Memory'),
                get_final_value(name + 'Memory')
            )

        if 'RepoName' in key:
            self.Image=get_sub_mapex(
                '${1M}.dkr.ecr.${AWS::Region}.amazonaws.com/${2M}:${EnvApp' + str(index) + 'Version}',
                ['EcrAccount', name + 'RepoName']
            )
        # use the same EnvApp version for all containers
        elif 'RepoName' in RP_cmm:
            self.Image=get_sub_mapex(
                '${1M}.dkr.ecr.${AWS::Region}.amazonaws.com/${2M}:${EnvApp1Version}',
                ['EcrAccount', 'RepoName']
            )
        elif 'Image' in RP_cmm:
            self.Image=get_final_value('Image')

        self.LogConfiguration=If(
            'LogConfiguration',
            ecs.LogConfiguration(
                LogDriver=get_final_value('LogDriver'),
                Options={
                    #'awslogs-group': get_final_value('AwsLogsGroup'),
                    #'awslogs-create-group': True,
                    'awslogs-group': Ref('LogsLogGroup'),
                    'awslogs-region': Ref('AWS::Region'),
                    'awslogs-stream-prefix': Ref('AWS::StackName')
                }
            ),
            Ref('AWS::NoValue')
        )

        if 'MountPoints' in key:
            self.MountPoints = [
                ECSMountPoint(n, key=k) for n, k in key['MountPoints'].iteritems()
            ]

        if 'Name' in key:
            self.Name = get_sub_mapex('${EnvRole}-${1M}', name + 'Name')
        else:
            self.Name = Ref('EnvRole')

        if 'ContainerPort' in key:
            PortMapping = ecs.PortMapping()
            auto_get_props(PortMapping, key, mapname=self.title)
            if 'HostPort' not in key:
                PortMapping.HostPort = If(
                    'NetworkModeAwsVpc',
                    get_final_value(name + 'ContainerPort'),
                    0
                )
            self.PortMappings = [PortMapping]


class ECSVolume(ecs.Volume):
    def __init__(self, title, key, **kwargs):
        super(ECSVolume, self).__init__(title, **kwargs)
        self.Name = self.title
        self.Host = ecs.Host(
            SourcePath=key['SourcePath']
        )


class ECSAwsvpcConfiguration(ecs.AwsvpcConfiguration):
    def setup(self):
        self.SecurityGroups = [
            GetAtt('SecurityGroupEcsService', 'GroupId'),
        ]
        self.Subnets = Split(',', get_exported_value('SubnetsPrivate'))


class ECRRepositoryLifecyclePolicy(ecr.LifecyclePolicy):
    def __init__(self, title, **kwargs):
        super(ECRRepositoryLifecyclePolicy, self).__init__(title, **kwargs)
        LifecyclePolicyText = {
            'rules': [
                {
                    'rulePriority': 1,
                    'description': 'Images are sorted on pushed_at_time (desc), images greater than specified count are expired.',
                    'selection': {
                        'tagStatus': 'any',
                        'countType': 'imageCountMoreThan',
                        'countNumber': 9500
                    },
                    'action': {
                        'type': 'expire'
                    }
                }
            ]
        }
        self.LifecyclePolicyText = json.dumps(LifecyclePolicyText)
        self.RegistryId = Ref('AWS::AccountId')


def ECRRepositoryPolicyStatementAccountPull(name):
    policy = {   
        'Action': [
            'ecr:GetDownloadUrlForLayer',
            'ecr:BatchGetImage',
            'ecr:BatchCheckLayerAvailability',
            'ecr:ListImages',
            'ecr:DescribeRepositories',
            'ecr:DescribeImages'
        ],
        'Effect': 'Allow',
        'Principal': {
            'AWS': [
                get_sub_mapex('arn:aws:iam::${1M}:root', name)
            ]
        },
        'Sid': 'AllowPull'
    }

    return policy


def ECRRepositoryPolicyStatementAccountPush(name):
    policy = {   
        'Action': [
            'ecr:PutImage',
            'ecr:InitiateLayerUpload',
            'ecr:UploadLayerPart',
            'ecr:CompleteLayerUpload',
        ],
        'Effect': 'Allow',
        'Principal': {
            'AWS': [
                get_sub_mapex('arn:aws:iam::${1M}:root', name)
            ]
        },
        'Sid': 'AllowPush'
    }

    return policy


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

class ECS_TaskDefinition(object):
    def __init__(self, key):
        Environments_Base = [
            ecs.Environment(
                Name='EnvRegion',
                Value=Ref('AWS::Region')
            ),
            ecs.Environment(
                Name='Env',
                Value=Ref('Env')
            ),
            ecs.Environment(
                Name='EnvAbbr',
                Value=Ref('EnvShort')
            ),
            ecs.Environment(
                Name='EnvBrand',
                Value=get_final_value('BrandDomain')
            ),
            ecs.Environment(
                Name='EnvRole',
                Value=Ref('EnvRole')
            ),
            ecs.Environment(
                Name='EnvStackName',
                Value=Ref('AWS::StackName')
            ),
            ecs.Environment(
                Name='EnvClusterStackName',
                Value=get_final_value('ClusterStack')
            ),
        ]

        # Parameters
        P_DockerLabelLastUpdate = Parameter('DockerLabelLastUpdate')
        P_DockerLabelLastUpdate.Description = 'Use to force redeploy'

        cfg.Parameters.extend([
            P_DockerLabelLastUpdate,
        ])

        Containers = []
        for n, v in RP_cmm['ContainerDefinitions'].iteritems():
            Environments = []
            MountPoints = []

            name = 'ContainerDefinitions' + str(n)  # Ex. ContainerDefinitions1

            # parameters
            # if ContainerDefinitions have RepoName use different EnvApp version
            if n == 1 or 'RepoName' in v:  
                nameenvapp = 'EnvApp' + str(n) + 'Version'  # Ex. EnvApp1Version

                EnvApp = Parameter(nameenvapp)
                EnvApp.Description = nameenvapp
                EnvApp.AllowedPattern = '^[a-zA-Z0-9-_.]*$'
                EnvApp.Default = '1'

                cfg.Parameters.append(EnvApp)

                # outputs
                o_EnvAppOut = Output(nameenvapp)
                o_EnvAppOut.Value = Ref(nameenvapp)

                # and use different output for RepoName
                if 'RepoName' in RP_cmm:
                    if 'RepoName' in v:
                        o_Repo = Output(name + 'RepoName')
                        o_Repo.Value = get_final_value(name + 'RepoName')
                    else:
                        o_Repo = Output('RepoName')
                        o_Repo.Value = get_final_value('RepoName')

                    cfg.Outputs.append(o_Repo)

                cfg.Outputs.extend([
                    o_EnvAppOut,
                ])

            # outputs
            EnvValue_Out_String = []
            EnvValue_Out_Map = {}
            if 'Envs' in v:
                for m, w in v['Envs'].iteritems():
                    envname = name + 'Envs' + str(m)
                    # parameters
                    EnvValue = Parameter(envname + 'Value')
                    EnvValue.Description = w['Name']  + ' - empty for default based on env/role'

                    cfg.Parameters.append(EnvValue)

                    Environment = ECSEnvironment(envname)
                    Environment.setup(key=w)
                    Environments.append(Environment)

                    # outputs
                    EnvValue_Out_String.append(w['Name'] + '=${' + w['Name'] + '}')
                    EnvValue_Out_Map.update({
                        w['Name']: Environment.Value
                    })

            o_EnvValueOut = Output(name + 'Envs')
            o_EnvValueOut.Value = Sub(','.join(EnvValue_Out_String), **EnvValue_Out_Map)

            cfg.Outputs.append(o_EnvValueOut)

            Environments.extend(Environments_Base)

            # parameters
            if 'Cpu' in v:
                p_Cpu = Parameter(name + 'Cpu')
                p_Cpu.Description = 'Cpu Share for containers - empty for default based on env/role'

                cfg.Parameters.append(p_Cpu)

            if 'Memory' in v:
                p_Memory = Parameter(name + 'Memory')
                p_Memory.Description = 'Memory hard limit for containers - empty for default based on env/role'

                cfg.Parameters.append(p_Memory)

            if 'MemoryReservation' in v:
                p_MemoryReservation = Parameter(name + 'MemoryReservation')
                p_MemoryReservation.Description = 'Memory soft limit for containers - empty for default based on env/role'

                cfg.Parameters.append(p_MemoryReservation)

            Container = ECSContainerDefinition(name)
            Container.setup(key=v, index=n)
            Container.Environment = Environments
            # Trick to force reload of Service using parameter
            Container.DockerLabels = If(
                'DockerLabelLastUpdateOverride',
                {'LastUpdate': Ref('DockerLabelLastUpdate')},
                Ref('AWS::NoValue'),
            )
            
            Containers.append(Container)

            # outputs
            Constraints_Out_String = []
            Constraints_Out_Map = {}

            if 'Cpu' in v:
                Constraints_Out_String.append('Cpu:${Cpu}')
                Constraints_Out_Map.update({
                    'Cpu': Container.Cpu
                })

            if 'Memory' in v:
                Constraints_Out_String.append('Memory:${Memory}')
                Constraints_Out_Map.update({
                    'Memory': Container.Memory
                })

            if 'MemoryReservation' in v:
                Constraints_Out_String.append('MemoryReservation:${MemoryReservation}')
                Constraints_Out_Map.update({
                    'MemoryReservation': Container.MemoryReservation
                })

            if Constraints_Out_String:
                o_Constraints = Output(name + 'Constraints')
                o_Constraints.Value = Sub(','.join(Constraints_Out_String), **Constraints_Out_Map)

                cfg.Outputs.append(o_Constraints)

        # Resources
        R_TaskDefinition = ECSTaskDefinition('TaskDefinition')
        R_TaskDefinition.setup()
        R_TaskDefinition.ContainerDefinitions = Containers

        if 'Volumes' in RP_cmm:
            Volumes = []
            for n, v in RP_cmm['Volumes'].iteritems():
                Volume = ECSVolume(n, key=v)
                Volumes.append(Volume)

            R_TaskDefinition.Volumes = Volumes

        cfg.Resources.extend([
            R_TaskDefinition,
        ])


class ECS_Service(object):
    def __init__(self, key):
        # Parameters
        P_MaximumPercent = Parameter('DeploymentConfigurationMaximumPercent')
        P_MaximumPercent.Description = 'DeploymentConfiguration MaximumPercent - empty for default based on env/role'

        P_MinimumHealthyPercent = Parameter('DeploymentConfigurationMinimumHealthyPercent')
        P_MinimumHealthyPercent.Description = 'DeploymentConfiguration MinimumHealthyPercent - empty for default based on env/role'

        cfg.Parameters.extend([
            P_MaximumPercent,
            P_MinimumHealthyPercent,
        ])

        # Resources
        R_SG = SecurityGroupEcsService('SecurityGroupEcsService')
        R_SG.setup()

        R_Service = ECSService('Service')

        if 'LoadBalancerApplication' in RP_cmm:
            if 'External' in RP_cmm['LoadBalancerApplication']:
                R_Service.setup(scheme='External')

                SGRule = SecurityGroupRuleEcsService()
                SGRule.setup(scheme='External')
                R_SG.SecurityGroupIngress.append(SGRule)

            if 'Internal' in RP_cmm['LoadBalancerApplication']:
                R_Service.setup(scheme='Internal')

                SGRule = SecurityGroupRuleEcsService()
                SGRule.setup(scheme='Internal')
                R_SG.SecurityGroupIngress.append(SGRule)
        else:
            R_Service.setup(scheme='')

        SecurityGroups = SG_SecurityGroupsECS().SecurityGroups
        NetworkConfiguration = ecs.NetworkConfiguration()
        NetworkConfiguration.AwsvpcConfiguration = ECSAwsvpcConfiguration()
        NetworkConfiguration.AwsvpcConfiguration.setup()
        NetworkConfiguration.AwsvpcConfiguration.SecurityGroups.extend(SecurityGroups)

        R_Service.NetworkConfiguration = If(
            'NetworkModeAwsVpc',
            NetworkConfiguration,
            Ref('AWS::NoValue')
        )

        cfg.Resources.extend([
            R_Service,
            R_SG,
        ])

        # Output
        try:
            O_MaximumPercent = Output('DeploymentConfigurationMaximumPercent')
            O_MaximumPercent.Value = R_Service.DeploymentConfiguration.MaximumPercent
    
            O_MinimumHealthyPercent = Output('DeploymentConfigurationMinimumHealthyPercent')
            O_MinimumHealthyPercent.Value = R_Service.DeploymentConfiguration.MinimumHealthyPercent
    
            cfg.Outputs.extend([
                O_MaximumPercent,
                O_MinimumHealthyPercent,
            ])
        except AttributeError:
            pass


class ECR_Repositories(object):
    def __init__(self, key):
        PolicyStatementAccounts = []
        for n, v in RP_cmm['EcrAccount'].iteritems():
            mapname = 'EcrAccount' + n  + 'Id'  # Ex. EcrAccountPrdId
            # conditions
            do_no_override(True)
            c_Account = {mapname: Not(
                Equals(get_final_value(mapname), 'None')
            )}

            cfg.Conditions.extend([
                c_Account,
            ])
            do_no_override(False)

            if 'Pull' in v['Policy']:
                PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPull(name=mapname)
                PolicyStatementAccounts.append(
                    If(
                        mapname,
                        PolicyStatementAccount,
                        Ref('AWS::NoValue')
                    )
                )

            if 'Push' in v['Policy']:
                PolicyStatementAccount = ECRRepositoryPolicyStatementAccountPush(name=mapname)
                PolicyStatementAccounts.append(
                    If(
                        mapname,
                        PolicyStatementAccount,
                        Ref('AWS::NoValue')
                    )
                )

        # Resources
        for n, v in RP_cmm[key].iteritems():
            Repo = ECRRepositories(key + n)  # Ex. RepositoryApiLocationHierarchy
            Repo.setup()
            Repo.RepositoryPolicyText['Statement'].extend(PolicyStatementAccounts)
            Repo.LifecyclePolicy = ECRRepositoryLifecyclePolicy('')

            cfg.Resources.append(Repo)


class ECS_Cluster(object):
    def __init__(self, key):
        # Resources
        R_Cluster = ecs.Cluster('Cluster')

        cfg.Resources.extend([
            R_Cluster,
        ])

# Need to stay as last lines
import_modules(globals())



import troposphere.logs as lgs

from shared import *

class LogsLogGroup(lgs.LogGroup):
    def __init__(self, title, **kwargs):
        super(LogsLogGroup, self).__init__(title, **kwargs)
        self.LogGroupName = get_final_value('LogGroupName')
        self.RetentionInDays = get_final_value('LogRetentionInDays')


# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class LGS_LogGroup(object):
    def __init__(self, key):
        R_Group = LogsLogGroup('LogsLogGroup')

        cfg.Resources.extend([
            R_Group,
        ])


# Need to stay as last lines
import_modules(globals())



import os
import importlib
import random
import string
import copy
import json
import logging
from pprint import pprint, pformat
import troposphere.ssm as ssm

from troposphere import (
    And,
    AWSHelperFn,
    AWSObject,
    AWSProperty,
    Base64,
    BaseAWSObject,
    Condition,
    Equals,
    Export,
    FindInMap,
    GetAtt,
    GetAZs,
    If,
    ImportValue,
    Join,
    Not,
    Or,
    Output,
    Parameter,
    Ref,
    Select,
    Split,
    Sub,
    Tags,
    Template,
)

from cfg import (
    mappedvalue,
    RP_cmm,
    MAX_SECURITY_GROUPS,
    SECURITY_GROUPS_DEFAULT,
    PARAMETERS_SKIP_OVERRIDE_CONDITION,
)

import cfg

# S - PARAMETERS #
class Parameter(Parameter):
    Type = 'String'
    Default = ''


class SSMParameter(ssm.Parameter):
    Type = 'String'
# E - PARAMETERS #


def stack_add_res():
    for v in cfg.Parameters:
        name = v.title
        # Automatically create override conditions for parameters
        if not name.startswith(PARAMETERS_SKIP_OVERRIDE_CONDITION):
            condition = {
                name + 'Override': Not(Equals(Ref(name), '' if name != 'InstanceType' else 'default'))
            }
            cfg.Conditions.append(condition)
            cfg.Parameters_Override.append(v)
        # End
        cfg.template.add_parameter(v)
    del cfg.Parameters[:]

    for v in cfg.Conditions:
        for name in v:
            cfg.template.add_condition(name, v[name])
    del cfg.Conditions[:]

    for v in cfg.Resources:
        cfg.template.add_resource(v)
    del cfg.Resources[:]

    for v in cfg.Outputs:
        cfg.template.add_output(v)
    del cfg.Outputs[:]


def do_no_override(action):
    if action is True:
        cfg.no_override = True
    else:
        cfg.no_override = False


def import_modules(gbl):
    for module in cfg.IMPORT_MODULES:
        mod = importlib.import_module('.%s' % module, package='awsibox')
        gbl.update(mod.__dict__)


def import_modules_old():
    for module in cfg.IMPORT_MODULES:
        mod = importlib.import_module(module)
        globals().update(mod.__dict__)


def get_final_value(
        param,
        ssm=False,
        condition=False,
        nocondition=False,
        nolist=False,
        inlist=False,
        split=False,
        issub=False,
        strout=False,
        mappedvalue=mappedvalue
):
    v = ''
    parameters = []

    # if param in mappedvalue means its value do not changes based on Env/Region so hardcode the value in json, ...
    if param in mappedvalue:
        mapped_value = mappedvalue[param]
        # check if value start with method and use eval to run code ... in list too
        if isinstance(mapped_value, list):
            mapped_value = [eval(r) if r.startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else r for r in mapped_value]
        if isinstance(mapped_value, str):
            mapped_value = eval(mapped_value.replace('\n', '')) if mapped_value.startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else mapped_value
    # ... otherway use mapping
    else:
        mapped_value = FindInMap(Ref('EnvShort'), Ref("AWS::Region"), param)

    if strout is True and isinstance(mapped_value, int):
        mapped_value = str(mapped_value)

    if nolist is True and isinstance(mapped_value, list):
        mapped_value = ','.join(mapped_value)

    if issub is False:
        override_value = If(param + 'Override', Ref(param), mapped_value)
    else:
        override_value = If(param + 'Override', Ref(param), Sub(mapped_value))

    parameters = cfg.Parameters_Override + cfg.Parameters

    if cfg.no_override is False and any(param == p.title for p in parameters):
        v = override_value
    # elif param in mappings['dev']['eu-west-1'] or param in mappedvalue:
    else:
        v = mapped_value
        # using ssm - DISABLED FOR NOW
        # v = mapped_value if ssm is True else GetAtt('SSMParameter' + param, 'Value')

    if condition:
        if condition is True:
            condname = param
        else:
            condname = condition

        v = If(
            condname,
            v,
            Ref('AWS::NoValue'),
        )
    elif nocondition:
        if nocondition is True:
            condname = param
        else:
            condname = nocondition

        v = If(
            condname,
            Ref('AWS::NoValue'),
            v,
        )

    if split:
        v = Select(split, Split(',', v))

    if inlist:
        v = Select(inlist, v)

    return v


def get_exported_value(param, stack=False, prefix=''):
    v = ''
    if stack:
        v = ImportValue(
            Sub(param + '-${' + stack + '}', **{stack: get_final_value(stack)})
        )
    elif not isinstance(param, str):
        v = ImportValue(
            Sub(prefix + '${ImportName}', **{'ImportName': param})
        )
    else:
        v = ImportValue(param)
 
    return v


def get_sub_mapex(substring, subvar, stack=False):
    submap = {}
    found = substring.find('${')

    while found != -1:
        posindex = found + 2
        myindex = substring[posindex]
        mytype = substring[posindex + 1]
        # M = Mapped, E = Exported
        if myindex.isdigit() and mytype in ['M', 'E']:
            listitem = subvar[int(myindex) - 1] if isinstance(subvar, list) else subvar
            stackitem = stack[int(myindex) - 1] if isinstance(stack, list) else stack
            if mytype == 'M':
                submap[listitem] = get_final_value(listitem)
            else:
                submap[listitem] = get_exported_value(listitem, stackitem)
            substring = substring.replace('${' + myindex + mytype + '}', '${' + listitem + '}')
        found = substring.find('${', posindex)

    v = Sub(substring, **submap)

    return v


def import_lambda(name):
    parent_dir_name = os.path.dirname(os.path.realpath(__file__))
    lambda_file = os.path.join(os.getcwd(), 'lib/lambdas/%s.code' % name)
    if not os.path.exists(lambda_file):
        lambda_file = os.path.join(parent_dir_name, 'lambdas/%s.code' % name)
    try:
        with open(lambda_file, 'r') as f:
            file_lines = []
            for x in f.readlines():
                if x.startswith(cfg.EVAL_FUNCTIONS_IN_CFG):
                    value = eval(x)
                elif x.startswith('STRING_EVAL_FUNCTIONS_IN_CFG'):
                    value = '"'
                else:
                    value = ''.join(x)
    
                file_lines.append(value)
                    
            #file_lines = [eval(x) if x.startswith(cfg.EVAL_FUNCTIONS_IN_CFG) else ''.join(x) for x in f.readlines()]
    
            return(file_lines)
    except IOError:
        logging.error('Lambda code %s not found' % name)
        exit(1)


def auto_get_props_recurse(obj, key, props, obj_propname, mapname, propname, root=None):
    prop_class = props[obj_propname][0]
    if isinstance(prop_class, type) and prop_class.__bases__[0].__name__ == 'AWSProperty':
        # If object already have that props, keep existing props
        if obj_propname in obj.properties:
            prop_obj = obj.properties[obj_propname]
        else:
            prop_obj = prop_class()
        auto_get_props(prop_obj, key=key[obj_propname], mapname=mapname + obj_propname, recurse=True, root=None if root is None else root[obj_propname])
        
        return prop_obj
    elif isinstance(prop_class, list) and isinstance(prop_class[0], type) and prop_class[0].__bases__[0].__name__ == 'AWSProperty':
        # If object props already is a list, keep existing list objects
        if obj_propname in obj.properties:
            prop_list = obj.properties[obj_propname]
        else:
            prop_list = []
        prop_class = props[obj_propname][0][0]
        for o, v in key[obj_propname].iteritems():
            prop_obj = prop_class()
            auto_get_props(prop_obj, key=v, mapname=mapname + obj_propname + str(o), recurse=True, root=None if root is None else root[obj_propname][o]) 
            prop_list.append(prop_obj)

            return prop_list

    return get_final_value(mapname + propname)


def auto_get_props_recurse(obj, key, props, obj_propname, mapname, propname, rootkey=None, rootname=None):
    prop_class = props[obj_propname][0]
    if isinstance(prop_class, type) and prop_class.__bases__[0].__name__ == 'AWSProperty':
        # If object already have that props, object class is the existing already defined object, so extend its property with auto_get_props
        if obj_propname in obj.properties:
            prop_obj = obj.properties[obj_propname]
        else:
            prop_obj = prop_class()
        auto_get_props(
            prop_obj,
            key=key[obj_propname],
            mapname=mapname + obj_propname,
            recurse=True,
            rootkey=rootkey[obj_propname] if rootkey else None,
            rootname=rootname + obj_propname if rootname else None,
        )
        
        return prop_obj
    elif isinstance(prop_class, list) and isinstance(prop_class[0], type) and prop_class[0].__bases__[0].__name__ == 'AWSProperty':
        # If object props already is a list, keep existing list objects
        # need to make a change, if object already exist, i must first iterate over old list with rootkey and rootname and then over new one 
        prop_list = []
        prop_class = props[obj_propname][0][0]
        # If rootkey is defined, first iterate over rootkey and execute auto_get_props passing rootkey and rootname, but check if element exist in key too,
        # in that case execute execute auto_get_props passing key and mapname...
        if rootkey:
            for o, v in rootkey[obj_propname].iteritems():
                name_o = str(o)
                rootname_o = rootname + obj_propname + name_o
                mapname_o = mapname + obj_propname + name_o
                prop_obj = prop_class()
                auto_get_props(
                    prop_obj,
                    key=v,
                    mapname=rootname_o,
                    recurse=True,
                    rootkey=v,
                    rootname=rootname_o,
                )
                if o in key[obj_propname]:
                    key_o = key[obj_propname][o]
                    auto_get_props(
                        prop_obj,
                        key=key_o,
                        mapname=mapname_o,
                        recurse=True,
                        rootkey=key_o,
                        rootname=rootname_o,
                    )
                prop_list.append(prop_obj)
        # ...then iterate over key, but check if element exist in rootkey too, in that case skip it (already included in previous rootkey iteration)
        # when calling auto_get_props rootkey and rootname can be setted to None, cause if element has not been skipped mean that it do not exist in rootkey
        # so there is no need to pass it (there cannot be a corrispective node in rootkey) 
        for o, v in key[obj_propname].iteritems():
            name_o = str(o)
            mapname_o = mapname + obj_propname + name_o
            prop_obj = prop_class()
            if rootkey and o in rootkey[obj_propname]:
                continue
            auto_get_props(
                prop_obj,
                key=v,
                mapname=mapname_o,
                recurse=True,
                rootkey=None,
                rootname=None,
            ) 
            prop_list.append(prop_obj)

        return prop_list

    return get_final_value(mapname + propname)


def auto_get_props(obj, key=RP_cmm, del_prefix='', mapname=None, recurse=False, rootkey=None, rootname=None, rootdict=None):
    # Allowed object properties
    props = vars(obj)['propnames']
    props.extend(vars(obj)['attributes'])
    # Object class
    classname = obj.__class__.__name__

    # build up mapname
    mapname = mapname if mapname is not None else obj.title
    if classname in ['Output', 'Parameter']:
        mapname = classname + mapname

    # iterate over props or key choosing the one with less objects
    use_key = True if len(props) > len(key) else None
    for propname in key if use_key else props:
        obj_propname = propname.replace(del_prefix, '') if use_key else del_prefix + propname
        if obj_propname in (props if use_key else key):
            if recurse and isinstance(key[obj_propname], dict):
                value = auto_get_props_recurse(obj, key, obj.props, obj_propname, mapname, propname, rootkey, rootname)
            # needed for lib/efs.py EFS_FileStorage SGIExtra - where is passed as key a new dictionary to parse for parameters
            elif rootdict:
                value = get_final_value(mapname + propname, mappedvalue=rootdict)
            else:
                value = get_final_value(mapname + propname)
            try:
                setattr(obj, obj_propname, value)
            except TypeError:
                pass


def auto_build_obj(obj, key, obj_list=cfg.Resources):
    props = vars(obj)['propnames']
    classname = obj.__class__
    for resname, resvalue in key.iteritems():
        final_obj = classname(resname)
        auto_get_props(final_obj, resvalue)

        obj_list.append(final_obj)


def gen_random_string():
    length = 16
    char_set = string.ascii_letters + string.digits
    if not hasattr(gen_random_string, "rng"):
        gen_random_string.rng = random.SystemRandom() # Create a static variable
    secret_string = ''.join([ gen_random_string.rng.choice(char_set) for _ in xrange(length) ])

    return secret_string


# Need to stay as last lines
import_modules(globals())



import troposphere.awslambda as lbd

from shared import *


class LambdaPermission(lbd.Permission):
    def setup(self):
        self.Action = 'lambda:InvokeFunction'


class LambdaPermissionEvent(LambdaPermission):
    def setup(self, key, source):
        super(LambdaPermissionEvent, self).setup()
        self.Principal = 'events.amazonaws.com'
        self.FunctionName = eval(key['Arn'])
        self.SourceArn = GetAtt(source, 'Arn')


class LambdaPermissionSNS(LambdaPermission):
    def setup(self, key, **kwargs):
        super(LambdaPermissionSNS, self).setup(**kwargs)
        auto_get_props(self, key)
        self.Principal = 'sns.amazonaws.com'
        self.FunctionName = eval(key['Endpoint'])
        self.SourceArn = eval(key['TopicArn'])


class LambdaPermissionApiGateway(LambdaPermission):
    def setup(self, name, source):
        super(LambdaPermissionApiGateway, self).setup()
        self.Principal = 'apigateway.amazonaws.com'
        self.FunctionName = name
        self.SourceArn = source


class LambdaVersion(lbd.Version):
    def __init__(self, title, name, **kwargs):
        super(LambdaVersion, self).__init__(title, **kwargs)
        self.FunctionName = Ref(name)


class LambdaFunction(lbd.Function):
    def setup(self, key, name):
        auto_get_props(self, key)
        import_name = key['ImportName'] if 'ImportName' in key else name
        self.Code = lbd.Code(
            ZipFile=Join('', import_lambda(import_name))
        )
        self.FunctionName = Sub('${AWS::StackName}-${EnvRole}-' + name)
        if 'Handler' not in key:
            self.Handler = 'index.lambda_handler'
        self.Role = GetAtt('RoleLambda' + name, 'Arn')

        if all(k in key for k in ['SecurityGroupIds', 'SubnetIds']):
            self.VpcConfig = lbd.VPCConfig('')
            auto_get_props(self.VpcConfig, key, mapname=self.title)

        if 'Variables' in key:
            self.Environment = lbd.Environment(
                Variables={
                    varname: get_final_value(self.title + 'Variables' + varname) for varname in key['Variables']
                }
            )
##
##

class LBD_Lambdas(object):
    def __init__(self, key):
        # Resources
        for n, v in RP_cmm[key].iteritems():
            resname = key + n
            r_Lambda = LambdaFunction(resname)
            r_Lambda.setup(key=v, name=n)

            if 'Enabled' in v:
                # conditions
                do_no_override(True)
                c_Lambda = {resname: Not(
                    Equals(get_final_value(resname + 'Enabled'), 'None')
                )}

                cfg.Conditions.append(c_Lambda)
                do_no_override(False)

                r_Lambda.Condition = resname
            if 'Version' in v:
                versionname = resname + 'Version'
                # parameters
                p_Version = Parameter(versionname)
                p_Version.Description = 'LambdaVersion change between A/B to force deploy new version'
                p_Version.AllowedValues = ['', 'A', 'B']
                p_Version.Default = ''

                cfg.Parameters.append(p_Version)

                # conditons
                do_no_override(True)
                c_VersionA = {versionname + 'A': Equals(
                    Ref(resname + 'Version'), 'A'
                )}
                
                c_VersionB = {versionname + 'B': Equals(
                    Ref(resname + 'Version'), 'B'
                )}
                c_Version = {versionname: Or(
                    Condition(versionname + 'A'),
                    Condition(versionname + 'B'),
                )}
                

                cfg.Conditions.extend([
                    c_VersionA,
                    c_VersionB,
                    c_Version,
                ])
                do_no_override(False)

                # resources
                r_VersionA = LambdaVersion(versionname + 'A', name=resname )
                r_VersionA.Condition = versionname + 'A'
                
                r_VersionB = LambdaVersion(versionname + 'B', name=resname )
                r_VersionB.Condition = versionname + 'B'

                cfg.Resources.extend([
                    r_VersionA,
                    r_VersionB,
                ])

                # output
                o_Version = Output(versionname)
                o_Version.Value = If(
                    versionname + 'A',
                    Ref(versionname + 'A'),
                    Ref(versionname + 'B')
                )
                o_Version.Condition = versionname
                
                cfg.Outputs.extend([
                    o_Version,
                ])

            r_Role = IAMRoleLambdaBase('Role' + resname)
            r_Role.setup(key=v)
            if hasattr(r_Lambda, 'Condition'):
            	r_Role.Condition = r_Lambda.Condition

            cfg.Resources.extend([
                r_Lambda,
                r_Role,
            ])

            if 'Export' in v and v['Export']:
                O_Lambda = Output(resname)
                O_Lambda.Value = GetAtt(resname, 'Arn')
                O_Lambda.Export = Export(resname)

                cfg.Outputs.append(O_Lambda)


# Need to stay as last lines
import_modules(globals())



import troposphere.iam as iam

from shared import *


# S - IAM #
class IAMUser(iam.User):
    def setup(self, key, name):
        self.Condition = self.title
        self.UserName = key['UserName']
        self.LoginProfile = iam.LoginProfile(
            Password=GetAtt('SSMParameterPassword' + name, 'Value'),
            PasswordResetRequired=True
        )


class IAMGroup(iam.Group):
    def setup(self, key, name):
        self.Condition = self.title
        self.GroupName = name


class IAMUserToGroupAddition(iam.UserToGroupAddition):
    def setup(self, key, name):
        self.Condition = name
        self.GroupName = Ref(name)


class IAMPolicy(iam.PolicyType):
    def setup(self, key, name):
        auto_get_props(self, key)
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }
        self.PolicyName = name


class IAMManagedPolicy(iam.ManagedPolicy):
    def setup(self, key, name):
        auto_get_props(self, key)
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }
        self.Description = key['Description']


class IAMPolicyBucketReplica(iam.PolicyType):
    def setup(self, bucket, key):
        name = self.title  # Ex. IAMPolicyReplicaBucketElasticSearch
        self.Condition = bucket + 'Replica'
        self.PolicyName = self.title
        self.PolicyDocument = {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': [
                        's3:GetReplicationConfiguration',
                        's3:ListBucket'
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        Sub('arn:aws:s3:::%s' % RP_cmm[bucket])
                    ]
                },
                {
                    'Action': [
                        's3:GetObjectVersion',
                        's3:GetObjectVersionAcl',
                        's3:GetObjectVersionTagging',
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        Sub('arn:aws:s3:::%s/*' % RP_cmm[bucket]) 
                    ]
                },
                {
                    'Action': [
                        's3:ReplicateObject',
                        's3:ReplicateDelete',
                        's3:ReplicateTags',
                        's3:ObjectOwnerOverrideToBucketOwner',
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        get_sub_mapex(
                            'arn:aws:s3:::${1M} + /*', '%sReplicaDstBucket' % bucket
                        ) if 'ReplicaDstBucket' in key else get_sub_mapex(
                            'arn:aws:s3:::${1M}-%s/*' % RP_cmm[bucket].replace('${AWS::Region}-', '', 1),
                            '%sReplicaDstRegion' % bucket,
                        )
                    ]
                }
            ]
        }
        self.Roles = [
            Ref('Role' + self.Condition)  # Ex. RoleBucketImagesReplica
        ]


class IAMInstanceProfile(iam.InstanceProfile):
    def setup(self):
        self.Path = '/'
        self.Roles=[
            Ref('RoleInstance')
        ]


class IAMRole(iam.Role):
    def setup(self, key):
        auto_get_props(self, key)
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Effect': 'Allow',
                'Principal': { 'Service': [key['Principal']] },
            }]
        }
        self.Path = '/'


class IAMRoleLambdaBase(iam.Role):
    def setup(self, key):
        self.Path = '/'
        self.ManagedPolicyArns = [
            'arn:aws:iam::aws:policy/AmazonEC2ReadOnlyAccess',
            'arn:aws:iam::aws:policy/service-role/AWSLambdaVPCAccessExecutionRole'
        ]
        if 'RoleManagedPolicyArns' in key:
            self.ManagedPolicyArns.extend(get_final_value('RoleManagedPolicyArns', mappedvalue=key))
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Principal': {'Service': [
                    'lambda.amazonaws.com',
                    'edgelambda.amazonaws.com' if 'AtEdge' in key else Ref('AWS::NoValue')
                ]},
                'Effect': 'Allow'
            }]
        }


class IAMRoleBucketReplica(iam.Role):
    def setup(self):
        self.Condition = self.title.replace('Role','')  # Ex. BucketPortalStaticReplica
        self.Path = '/'
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Effect': 'Allow',
                'Principal': {'Service': ['s3.amazonaws.com']},
            }],
            'Version': '2012-10-17'
        }


# E - IAM #

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

def IAMPolicyStatement(key):
    Statement = {
        'Effect': key['Effect'] if 'Effect' in key else 'Allow',
    }

    for k in ['Action', 'NotAction', 'Resource', 'NotResource', 'Condition']:
        if k in key:
            Statement.update({k: get_final_value(k, mappedvalue=key)})

    return Statement

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class IAM_Users(object):
    def __init__(self, key):
        for n, v in RP_cmm[key].iteritems():
            resname = key + n  # Ex. IAMUserPincoPalla

            # parameters
            p_Password = Parameter('PasswordBase' + n)
            p_Password.Description = 'Base Password, must be changed at first login'
            p_Password.Default = ''
            p_Password.NoEcho = True

            cfg.Parameters.append(p_Password)

            # conditions
            do_no_override(True)
            c_User = {resname: Not(
                Equals(get_final_value(resname + 'Enabled'), 'None')
            )}
            c_Role = {resname + 'RoleAccount': Not(
                Equals(get_final_value(resname + 'RoleAccount'), 'None')
            )}

            cfg.Conditions.extend([
                c_User,
                c_Role,
            ])
            do_no_override(False)

            ManagedPolicyArns = []
            RoleGroups = []
            if 'RoleGroups' in v:
                for m, w in v['RoleGroups'].iteritems():
                    condname = resname + 'RoleGroups' + m
                    # conditions
                    do_no_override(True)
                    c_RoleGroup = {condname: Not(
                        Equals(get_final_value(condname), 'None')
                    )}
    
                    cfg.Conditions.append(c_RoleGroup)
                    do_no_override(False)
    
                    # resources
                    RoleGroups.append(
                        If(
                            condname,
                            m,
                            Ref('AWS::NoValue')
                        )
                    )
    
                    if m in RP_cmm['IAMGroup']:
                        for p in RP_cmm['IAMGroup'][m]['ManagedPolicyArns']:
                            ManagedPolicyArns.append(
                                If(
                                    condname,
                                    Ref('IAMPolicy' + p),
                                    Ref('AWS::NoValue')
                                )
                            )

            # resources
            r_Role = IAMRole('IAMRole' + n)
            r_Role.setup(key={'Principal': ''})
            r_Role.Condition = resname + 'RoleAccount'
            r_Role.RoleName = v['UserName']
            r_Role.MaxSessionDuration = 43200
            r_Role.ManagedPolicyArns = ManagedPolicyArns
            AssumeRoleStatement = r_Role.AssumeRolePolicyDocument['Statement'][0]
            AssumeRoleStatement['Condition'] = {
                'StringEquals': {'aws:username': v['UserName']}
            }
            AssumeRoleStatement['Principal'] = {
                'AWS': Sub('arn:aws:iam::${IdAccount}:root', **{'IdAccount': get_final_value(resname + 'RoleAccount')})
            }

            r_User = IAMUser(resname)
            r_User.setup(key=v, name=n)
            r_User.Groups = RoleGroups

            r_SSMParameter = SSMParameter('SSMParameterPassword' + n)
            r_SSMParameter.Condition = resname
            r_SSMParameter.Name = Sub('/iam/PasswordBase/${UserName}', **{'UserName': get_final_value(resname + 'UserName')})
            r_SSMParameter.Value = Ref('PasswordBase' + n)
            r_SSMParameter.AllowedPattern = '^[^ ]{16,}$'

            cfg.Resources.extend([
                r_User,
                r_Role,
                r_SSMParameter,
            ])


class IAM_Groups(object):
    def __init__(self, key):
        for n, v in RP_cmm[key].iteritems():
            resname = key + n  # Ex. IAMGroupBase

            # conditions
            do_no_override(True)
            c_Group = {resname: Not(
                Equals(get_final_value(resname + 'Enabled'), 'None')
            )}
            cfg.Conditions.append(c_Group)
            do_no_override(False)

            ManagedPolicyArns = []
            for m in v['ManagedPolicyArns']:
                if m.startswith('arn'):
                    ManagedPolicyArns.append(m)
                else:
                    ManagedPolicyArns.append(Ref('IAMPolicy' + m))

            Users = []
            for m, w in v['User'].iteritems():
                condname = resname + 'User' + str(m)
                # conditions
                do_no_override(True)
                c_User = {condname: Not(
                    Equals(get_final_value(condname), 'None') 
                )}
                cfg.Conditions.append(c_User)
                do_no_override(False)

                Users.append(
                    If(
                        condname,
                        Ref('IAMUser' + m) if m in RP_cmm['IAMUser'] else get_final_value(condname),
                        Ref('AWS::NoValue')
                    )
                )

            # resources
            r_Group = IAMGroup(resname)
            r_Group.setup(key=v, name=n)
            r_Group.ManagedPolicyArns = ManagedPolicyArns

            r_GroupAdd = IAMUserToGroupAddition('IAMUserToGroupAddition' + n)
            r_GroupAdd.setup(key=v, name=resname)
            r_GroupAdd.Users = Users

            cfg.Resources.extend([
                r_Group,
                r_GroupAdd,
            ])


class IAM_Policies(object):
    def __init__(self, key):
        # Resources
        for n, v in RP_cmm[key].iteritems():
            resname = key + n  # Ex. IAMPolicyLambdaR53RecordInstanceId
            Statement = []
            for m, w  in v['Statement'].iteritems():
                Statement.append(IAMPolicyStatement(w))

            if 'Type' in v and v['Type'] == 'Managed':
                r_Policy = IAMManagedPolicy(resname)
            else:
                r_Policy = IAMPolicy(resname)
            r_Policy.setup(key=v, name=n)
            r_Policy.PolicyDocument['Statement'] = Statement

            cfg.Resources.append(r_Policy)

            # outputs
            if 'Export' in v:
                o_Policy = Output(resname)
                o_Policy.Value = Ref(resname)
                o_Policy.Export = Export(resname)

                cfg.Outputs.append(o_Policy)


class IAM_Roles(object):
    def __init__(self, key):
        # Resources
        for n, v in RP_cmm[key].iteritems():
            resname = key + n  # Ex. RoleECSService
            r_Role = IAMRole(resname)
            r_Role.setup(key=v)

            cfg.Resources.append(r_Role)

            # outputs
            if 'Export' in v:
                o_Role = Output(resname)
                o_Role.Value = GetAtt(resname, 'Arn')
                o_Role.Export = Export(resname)

                cfg.Outputs.append(o_Role)

# Need to stay as last lines
import_modules(globals())



import troposphere.ec2 as ec2

from shared import *


# S - SECURITY GROUP #
class SecurityGroup(ec2.SecurityGroup):
    def setup(self):
        self.VpcId = get_exported_value('VpcId')


class SecurityGroupInstanceRules(SecurityGroup):
    def setup(self):
        super(SecurityGroupInstanceRules, self).setup()
        self.GroupDescription = 'Enable Access from LoadBalancer to Instances and between Instances'


class SecurityGroupLoadBalancer(SecurityGroup):
    def setup(self):
        super(SecurityGroupLoadBalancer, self).setup()
        self.GroupDescription = 'Enable access to LoadBalancer'


class SecurityGroupEcsService(SecurityGroup):
    def setup(self):
        super(SecurityGroupEcsService, self).setup()
        self.Condition = 'NetworkModeAwsVpc'
        self.GroupDescription = 'Enable access to Service'
        self.SecurityGroupIngress = []


class SecurityGroupIngress(ec2.SecurityGroupIngress):
    IpProtocol = 'tcp'


class SecurityGroupIngressInstanceELBPorts(SecurityGroupIngress):
    def setup(self, listener):
        name = self.title  # Ex. SecurityGroupIngressListeners1
        self.FromPort = get_final_value(listener + 'LoadBalancerPort')
        self.ToPort = get_final_value(listener + 'LoadBalancerPort')
        self.SourceSecurityGroupId = Ref('SecurityGroupLoadBalancer')
        self.GroupId = GetAtt('SecurityGroupInstancesRules', 'GroupId')


class SecurityGroupsIngressEcs(SecurityGroupIngress):
    def setup(self, proto, scheme):
        self.Condition = self.title
        self.GroupId = get_exported_value('SecurityGroupLoadBalancerApplication' + scheme)
        self.FromPort = get_final_value('ListenerLoadBalancer' + proto + 'Port')
        self.ToPort = self.FromPort

# E - SECURITY GROUP #


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

# S - SECURITY GROUP #
class SecurityGroupRule(ec2.SecurityGroupRule):
    IpProtocol = 'tcp'


class SecurityGroupRuleELBPorts(SecurityGroupRule):
    def setup(self):
        name = self.title  # Ex. Listeners1
        self.FromPort = get_final_value(name + 'LoadBalancerPort')
        self.ToPort = get_final_value(name + 'LoadBalancerPort')


class SecurityGroupRuleNamePorts(SecurityGroupRule):
    def setup(self):
        name = self.title  # Ex. 3306
        self.FromPort = name
        self.ToPort = name


class SecurityGroupRuleEcsService(SecurityGroupRule):
    def setup(self, scheme):
        self.SourceSecurityGroupId = get_exported_value('SecurityGroupLoadBalancerApplication' + scheme)
        self.FromPort = get_final_value('ContainerDefinitions1ContainerPort')
        self.ToPort = self.FromPort

# E - SECURITY GROUP #

# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

class SG_SecurityGroupsExtra(object):
    def __init__(self, Out_String, Out_Map):
        # Parameters
        P_SecurityGroups = Parameter('SecurityGroups')
        P_SecurityGroups.Description = 'SecurityGroups List Extra - ' + SECURITY_GROUPS_DEFAULT + ' for default based on env/role'
        P_SecurityGroups.AllowedPattern = '^(\w*,\w*){' + str(MAX_SECURITY_GROUPS - 1) + '}$'
        P_SecurityGroups.Default = SECURITY_GROUPS_DEFAULT

        cfg.Parameters.extend([
            P_SecurityGroups,
        ])

        # Conditions
        do_no_override(True)
        C_Override = {'SecurityGroupsOverride': Not(
            Equals(Select('0', Split(',', Ref('SecurityGroups'))), '')
        )}

        cfg.Conditions.extend([
            C_Override,
        ])
        do_no_override(False)

        SecurityGroups = []

        for n in range(MAX_SECURITY_GROUPS):
            name = 'SecurityGroup' + str(n)  # Ex SecurityGroup1
            value = Select(n, Split(',', get_final_value('SecurityGroups')))  # Ex. ElasticSearch
            outnamename = 'SecurityGroupName' + str(n)  # Ex. SecurityGroupName1
            outvaluename = 'SecurityGroupValue' + str(n)  # Ex. SecurityGroupValue1

            # conditions
            do_no_override(True)
            c_OverrideCondition = {name: Not(
                Or(
                    And(
                        Condition('SecurityGroupsOverride'),
                        Equals(Select(n, Split(',', Ref('SecurityGroups'))), 'None')
                    ),
                    And(
                        Not(Condition('SecurityGroupsOverride')),
                        Equals(Select(n, Split(',', get_final_value('SecurityGroups'))), 'None')
                    )
                )
            )}
            cfg.Conditions.append(c_OverrideCondition)
            do_no_override(False)

            SecurityGroups.append(If(
                name,
                get_exported_value(value, prefix='SecurityGroup'),
                Ref('AWS::NoValue')
            ))

            # outputs
            Out_String.append('${' + outnamename + '}=${' + outvaluename + '}')
            Out_Map.update({
                outnamename: value,
                outvaluename: If(
                    name,
                    get_exported_value(value, prefix='SecurityGroup'),
                    'None'
                )
            })

        # Outputs
        O_SecurityGroups = Output('SecurityGroups')
        O_SecurityGroups.Value = Sub(','.join(Out_String), **Out_Map)

        cfg.Outputs.extend([
            O_SecurityGroups,
        ])

        self.O_SecurityGroups = O_SecurityGroups

        self.SecurityGroups = SecurityGroups


class SG_SecurityGroupIngressesExtra(object):
    def __init__(self, key):
        for n, v in RP_cmm[key].iteritems():
            resname = key + n
            r_SGI = SecurityGroupIngress(resname)
            auto_get_props(r_SGI, v)

            cfg.Resources.append(r_SGI)


class SG_SecurityGroupsEC2(object):
    def __init__(self):
        Out_String = ['Rules=${SecurityGroupInstancesRules}']
        Out_Map = {
            'SecurityGroupInstancesRules': {
                'Ref': 'SecurityGroupInstancesRules'
            },
        }


        # Resources
        SG_Extra = SG_SecurityGroupsExtra(Out_String, Out_Map)

        R_SGInstance = SecurityGroupInstanceRules('SecurityGroupInstancesRules')
        R_SGInstance.setup()

        cfg.Resources.extend([
            R_SGInstance,
        ])

        self.SecurityGroups = SG_Extra.SecurityGroups


class SG_SecurityGroupsECS(object):
    def __init__(self):
        Out_String = ['Service=${SecurityGroupEcsService}']
        Out_Map = {
            'SecurityGroupEcsService': {
                'Ref': 'SecurityGroupEcsService'
            },
        }

        # Resources
        SG_Extra = SG_SecurityGroupsExtra(Out_String, Out_Map)

        SG_Extra.O_SecurityGroups.Condition = 'NetworkModeAwsVpc'

        self.SecurityGroups = SG_Extra.SecurityGroups


class SG_SecurityGroupsTSK(object):
    def __init__(self):
        Out_String = []
        Out_Map = {}

        # Resources
        SG_Extra = SG_SecurityGroupsExtra(Out_String, Out_Map)

        SG_Extra.O_SecurityGroups.Condition = 'NetworkModeAwsVpc'

        self.SecurityGroups = SG_Extra.SecurityGroups


class SG_SecurityGroupRES(object):
    def __init__(self, key):
        for n, v in RP_cmm['SecurityGroupBase'].iteritems():
            name = 'SecurityGroup%s' % n  # Ex. SecurityGroupElasticSearchClient
            outname = name if n != 'ElasticSearchClient' else 'SecurityGroup' + n.replace('Client', '')  # Ex. SecurityGroupElasticSearch
            # resources
            r_Base = SecurityGroup(name)
            r_Base.setup()
            r_Base.GroupDescription = get_final_value('SecurityGroupBase' + n)

            cfg.Resources.append(r_Base)

            # outputs
            o_Base = Output(outname)
            o_Base.Value = GetAtt(name, 'GroupId')
            o_Base.Export = Export(outname)

            cfg.Outputs.append(o_Base)

        Securitygroup_Rules = []
        for n, v in RP_cmm['AllowedIp'].iteritems():
            name = 'AllowedIp%s' % n
            # conditions
            do_no_override(True)
            cfg.Conditions.append({
                name: Not(Equals(get_final_value(name + 'Enabled'), 'None'))
            })
            do_no_override(False)

            # resources
            Rule = SecurityGroupRule()
            Rule.FromPort = '22'
            Rule.ToPort = '22'
            Rule.CidrIp=get_final_value(name + 'Ip')
            
            Securitygroup_Rules.append(
                If(
                    name,
                    Rule,
                    Ref('AWS::NoValue')
                )
            )

        Securitygroup_Rules.append(
            ec2.SecurityGroupRule(
                CidrIp='0.0.0.0/0',
                FromPort='8',
                IpProtocol='icmp',
                ToPort='-1'
            )
        )      

        # Resources
        R_BaseInstance = SecurityGroup('SecurityGroupBaseInstance')
        R_BaseInstance.setup()
        R_BaseInstance.GroupDescription = 'Enable ICMP Ping [ALL] and SSH [192.168.36.0/23]'
        # R_BaseInstance.GroupDescription='Enable ICMP Ping [ALL] and SSH [' + ','.join(securitygroups_description['resources-env']) + ']'
        R_BaseInstance.SecurityGroupIngress = Securitygroup_Rules

        cfg.Resources.extend([
            R_BaseInstance,
        ])

        # Outputs
        O_BaseInstance = Output('SecurityGroupBaseInstance')
        O_BaseInstance.Value = GetAtt('SecurityGroupBaseInstance', 'GroupId')
        O_BaseInstance.Export = Export('SecurityGroupBaseInstance')

        cfg.Outputs.extend([
            O_BaseInstance,
        ])


class SG_SecurityGroupIngressesExtraService(object):
    def __init__(self, key, service):  # Ex service=RDS
        Securitygroup_Rules = []
        for n, v in RP_cmm[key].iteritems():
            name = str(v['FromPort'])  # Ex 3306
            for i in RP_cmm['AllowedIp']:
                ipname = 'AllowedIp' + str(i)  # Ex. AllowedIp1
                condnameprivate = 'SecurityGroupRulePrivatePort' + name + ipname  # Ex. SecurityGroupRulePrivatePorts3306AllowedIp1

                # conditions
                do_no_override(True)
                c_AllowedIp = {ipname: Not(
                    Equals(get_final_value(ipname + 'Enabled'), 'None')
                )}

                c_SGRulePrivate = {condnameprivate: And(
                    Condition(ipname),
                    Equals(get_final_value(service + 'Access'), 'Private')
                )}

                cfg.Conditions.extend([
                    c_AllowedIp,
                    c_SGRulePrivate,
                ])
                do_no_override(False)

                SGRule = SecurityGroupRuleNamePorts(name)
                SGRule.setup()
                SGRule.CidrIp = get_final_value(ipname + 'Ip')

                Securitygroup_Rules.append(
                    If( 
                        condnameprivate,
                        SGRule,
                        Ref('AWS::NoValue'),
                    )
                )

            SGRule = SecurityGroupRuleNamePorts(name)
            SGRule.setup()
            SGRule.CidrIp = '0.0.0.0/0'

            Securitygroup_Rules.append(
                If(
                    service + 'Public',
                    SGRule,
                    Ref('AWS::NoValue'),
                )
            )

        # Resources
        SG_SecurityGroupIngressesExtra(key=key)

        R_SG = SecurityGroup('SecurityGroup' + service)
        R_SG.setup()
        R_SG.GroupDescription = 'Enable access to ' + service
        R_SG.SecurityGroupIngress = Securitygroup_Rules

        cfg.Resources.extend([
            R_SG,
        ])

        # Outputs
        O_SG = Output('SecurityGroup')
        O_SG.Value = GetAtt('SecurityGroup' + service, 'GroupId')

        cfg.Outputs.extend([
            O_SG,
        ])

class SG_SecurityGroupIngressesExtraRDS(SG_SecurityGroupIngressesExtraService):
    def __init__(self, **kwargs):
        super(SG_SecurityGroupIngressesExtraRDS, self).__init__(service='RDS', **kwargs)


class SG_SecurityGroupIngressesExtraCCH(SG_SecurityGroupIngressesExtraService):
    def __init__(self, **kwargs):
        super(SG_SecurityGroupIngressesExtraCCH, self).__init__(service='CCH', **kwargs)


# Need to stay as last lines
import_modules(globals())



import troposphere.apigateway as agw

from shared import *
from collections import OrderedDict, Mapping

class ApiGatewayAccount(agw.Account):
    def setup(self):
        self.CloudWatchRoleArn = get_exported_value('RoleApiGatewayCloudWatch')


class ApiGatewayResource(agw.Resource):
    def setup(self, key, stage):
        mapname = 'ApiGatewayStage' + stage + 'ApiGatewayResource'
        stagekey = RP_cmm['ApiGatewayStage'][stage]['ApiGatewayResource']
        auto_get_props(self, key)
        self.RestApiId = Ref('ApiGatewayRestApi')
        auto_get_props(self, stagekey, mapname=mapname, recurse=True, rootkey=key, rootname=self.title)


class ApiGatewayMethod(agw.Method):
    def setup(self, key, basename, name, stage):
        mapname = 'ApiGatewayStage' + stage + 'ApiGatewayResource' + basename + 'Method' + name
        stagekey = RP_cmm['ApiGatewayStage'][stage]['ApiGatewayResource'][basename]['Method'][name] 
        auto_get_props(self, key, recurse=True)
        self.RestApiId = Ref('ApiGatewayRestApi')
        self.ResourceId = Ref('ApiGatewayResource' + basename)
        auto_get_props(self, stagekey, mapname=mapname, recurse=True, rootkey=key, rootname=self.title)
        # If Uri is a lambda get same stage version lambda name Ex. LambdaPaidApiv1
        if ':lambda:' in self.Integration.Uri:
            st = self.Integration.Uri
            l = st.rfind('.')
            self.Integration.Uri = Sub(st[0:l] + stage + st[l:])


class ApiGatewayStage(agw.Stage):
    def setup(self, name, key):
        self.StageName = name
        auto_get_props(self, key, recurse=True)
        self.RestApiId = Ref('ApiGatewayRestApi')
        self.DeploymentId = Ref('ApiGatewayDeployment' + name)


class ApiGatewayDeployment(agw.Deployment):
    def setup(self, name, key):
        lastresource = next(reversed(RP_cmm['ApiGatewayResource']))
        lastmethod = next(reversed(RP_cmm['ApiGatewayResource'][lastresource]['Method']))
        self.DependsOn = 'ApiGatewayResource' + lastresource + 'Method' + lastmethod
        self.Description = Ref('Deployment' + name + 'Description')
        self.RestApiId = Ref('ApiGatewayRestApi')


class AGW_Stages(object):
    def __init__(self, key):
        for n, v in RP_cmm[key].iteritems():
            # parameters
            p_DeploymentDescription = Parameter('Deployment' + n + 'Description')
            p_DeploymentDescription.Description = 'Deployment' + n + ' Description'
            p_DeploymentDescription.Default = n

            cfg.Parameters.extend([
                p_DeploymentDescription,
            ])

            # resources
            resname = key + n
            r_Stage = ApiGatewayStage(resname)
            r_Stage.setup(name=n, key=v)

            r_Deployment = ApiGatewayDeployment('ApiGatewayDeployment' + n)
            r_Deployment.setup(name=n, key=v)

            cfg.Resources.extend([
                r_Stage,
                r_Deployment,
            ])
       

class AGW_RestApi(object):
    def __init__(self, key):
        # Resources
        R_RestApi = agw.RestApi('ApiGatewayRestApi')
        auto_get_props(R_RestApi, RP_cmm, mapname='')

        R_Account = ApiGatewayAccount('ApiGatewayAccount')
        R_Account.setup()

        cfg.Resources.extend([
            R_RestApi,
            R_Account,
        ])

        for n, v in RP_cmm[key].iteritems():
            resname = key + n
            agw_stage = RP_cmm['Stage']
            r_Resource = ApiGatewayResource(resname)
            r_Resource.setup(key=v, stage=agw_stage)

            for m, w in v['Method'].iteritems():
                r_Method = ApiGatewayMethod(resname + 'Method' + m)
                r_Method.setup(key=w, basename=n, name=m, stage=agw_stage)

                cfg.Resources.extend([
                    r_Resource,
                    r_Method,
                ])

        for l, w in RP_cmm['Lambda'].iteritems():
            r_LambdaPermission = LambdaPermissionApiGateway('LambdaPermission' + l)
            r_LambdaPermission.setup(
                name=Ref('Lambda' + l),
                source=Sub('arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:${ApiGatewayRestApi}/*/*/*')
            )

            cfg.Resources.append(r_LambdaPermission)



# Need to stay as last lines
import_modules(globals())



import troposphere.cloudformation as cfm
from shared import *


def cfn_ecs_cluster():
    init_args = {}

    init_args.update({
        'REPOSITORIES': cfm.InitConfig(
            commands=If('GPUInstance', {
                '10-epel-repo': {
                    'command': 'amazon-linux-extras install -y epel',
                    'test': '! test -e /etc/yum.repos.d/epel.repo'
                },
                #'cuda-repo': {
                #    'command': 'yum install -y http://developer.download.nvidia.com/compute/cuda/repos/rhel7/x86_64/cuda-repo-rhel7-10.0.130-1.x86_64.rpm',
                #    'test': 'test ! -e /etc/yum.repos.d/cuda.repo'
                #},
                'nvidia-docker': {
                    'command': 'curl -s -L https://nvidia.github.io/nvidia-docker/amzn2/nvidia-docker.repo | tee /etc/yum.repos.d/nvidia-docker.repo',
                    'test': 'test ! -e /etc/yum.repos.d/nvidia-docker.repo'
                },
            }, Ref('AWS::NoValue')),
            files=If('GPUInstance', {
                '/etc/yum.repos.d/epel-nvidia.repo': {
                    'content': Join('\n', [
                        '[epel-nvidia]',
                        'name=negativo17 - Nvidia',
                        'baseurl=https://negativo17.org/repos/nvidia/epel-7/$basearch/',
                        'enabled=1',
                        'skip_if_unavailable=1',
                        'gpgcheck=1',
                        'gpgkey=https://negativo17.org/repos/RPM-GPG-KEY-slaanesh',
                        'enabled_metadata=1',
                        'metadata_expire=6h',
                        'type=rpm-md',
                        'repo_gpgcheck=0'
                    ])
                }
            }, Ref('AWS::NoValue'))
        ),
        'PACKAGES': cfm.InitConfig(
            packages={
                'yum': {
                    'nfs-utils': [],
                    'make': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'automake': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'gcc': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'gcc-c++': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'kernel-devel': If('GPUInstance', [], Ref('AWS::NoValue')),
                    #'cuda': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'dkms-nvidia': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'nvidia-driver-cuda': If('GPUInstance', [], Ref('AWS::NoValue')),
                    'nvidia-docker2': If('GPUInstance', [], Ref('AWS::NoValue')),
                }
            },
            files={
                '/etc/profile.d/ibox_env_ext.sh': If('SpotASG', {
                    'content': Join('', [
                        '#setup ibox extra environment variables\n',
                        'SNSTopicASGSpot=', Ref('SNSTopicASGSpot'), '\n',
                    ])
                }, Ref('AWS::NoValue')) if 'SpotASG' in RP_cmm else Ref('AWS::NoValue'),
                '/etc/systemd/system/spot-instance-drainer.service': If('SpotPrice', {
                    'content': Join('\n', [
                        '[Unit]',
                        'Description=Spot Instance Drainer',
                        '[Service]',
                        'Type=simple',
                        'EnvironmentFile=-/etc/profile.d/ibox_env_ext.sh',
                        'ExecStart=/usr/local/bin/spot-instance-drainer',
                        'ExecStop=/usr/bin/killall spot-instance-drainer',
                        'Restart=on-failure',
                        '[Install]',
                        'WantedBy=multi-user.target',
                    ]),
                }, Ref('AWS::NoValue')),
                '/usr/local/bin/spot-instance-drainer': If('SpotPrice', {
                    'source': Sub(
                        'https://' + get_final_value('BucketAppRepository')  + '.s3-${AWS::Region}.amazonaws.com/ibox-tools/spot-instance-drainer'
                    ),
                    'mode': 755
                }, Ref('AWS::NoValue')),
            }
        ),
        'SERVICES': cfm.InitConfig(
            files={
                '/etc/ecs/ecs.config': {
                    'content': Join('\n', [
                        Sub('ECS_CLUSTER=${Cluster}'),
                        If('GPUInstance', 'ECS_DISABLE_PRIVILEGED=false', Ref('AWS::NoValue')),
                    ])
                },
                '/etc/docker/daemon.json': If('GPUInstance', {
                    'content': Join('\n', [
                        '{',
                        '  "default-runtime": "nvidia",',
                        '  "runtimes": {',
                        '    "nvidia": {',
                        '      "path": "/usr/bin/nvidia-container-runtime",',
                        '      "runtimeArgs": []',
                        '    }',
                        '  }',
                        '}',
                    ])
                }, Ref('AWS::NoValue'))
            },
            commands={
                '01-kern-modules': If('GPUInstance', {
                    'command': 'rmmod nvidia_modeset;rmmod nvidia_uvm;rmmod nvidia;modprobe nvidia;modprobe nvidia_uvm;true'
                }, Ref('AWS::NoValue')),
                '02-restart-docker': If('GPUInstance', {
                    'command': 'pkill -SIGHUP dockerd'
                }, Ref('AWS::NoValue')),
                '03-reload-systemd': If('SpotPrice', {
                    'command': 'systemctl daemon-reload'
                }, Ref('AWS::NoValue')),
            },
            services=If('SpotPrice', {
                'sysvinit': {
                    'spot-instance-drainer': {
                        'enabled': 'false',
                        'ensureRunning': 'true'
                    }
                }
            }, Ref('AWS::NoValue'))
        )
    })

    return init_args



import yaml
import yaml.constructor
import sys
import os
import copy
import json
import logging
from collections import OrderedDict, Mapping
from pprint import pprint, pformat

import cfg


def show_odict(odict):
    print(json.dumps(odict, indent=4))


# We also need to load mappings in order
# Based on https://gist.github.com/844388
def construct_ordereddict(loader, node):
    data = OrderedDict()
    yield data
    value = construct_mapping(loader, node)
    data.update(value)


def construct_mapping(self, node, deep=False):
    if isinstance(node, yaml.MappingNode):
        self.flatten_mapping(node)
    else:
        raise yaml.constructor.ConstructorError(
            None, None, 'expected a mapping node, but found %s' % node.id, node.start_mark)

    mapping = OrderedDict()
    for key_node, value_node in node.value:
        key = self.construct_object(key_node, deep=deep)
        try:
            hash(key)
        except TypeError, exc:
            raise yaml.constructor.ConstructorError(
                'while constructing a mapping', node.start_mark,
                'found unacceptable key (%s)' % exc, key_node.start_mark)
        value = self.construct_object(value_node, deep=deep)
        mapping[key] = value
    return mapping


class Loader(yaml.Loader):
    def __init__(self, stream):
        # This way for include relative to file with include statement
        self._root_current = os.path.split(stream.name)[0]
        # This way for include BASE relative on BASE dir
        self._root_base = os.path.join(CFG_FILE_INT, 'BASE')
        # This way for include relative on BASE EXT dir
        self._root_brand_base = os.path.join(CFG_FILE_EXT, 'BASE')
        # This way for include relative on brand EXT dir
        self._root_brand_ext = os.path.join(CFG_FILE_EXT, brand)
        self.stream = stream
        super(Loader, self).__init__(stream)
        Loader.add_constructor('!include', Loader.include)
        Loader.add_constructor('!import',  Loader.include)

    def include(self, node):
        if   isinstance(node, yaml.ScalarNode):
            return self.extractFile(self.construct_scalar(node), self._root_current)

        elif isinstance(node, yaml.SequenceNode):
            result = []
            for filename in self.construct_sequence(node):
                # if including Ext cfg try to include BASE file from CFG_FILE_INT too
                if CFG_FILE_EXT in self.stream.name:
                    result.append(self.extractFile(filename, self._root_base))
                result.append(self.extractFile(filename, self._root_current))
                # try to include BASE ext too
                result.append(self.extractFile(filename, self._root_brand_base))
                # try to include brand ext too
                result.append(self.extractFile(filename, self._root_brand_ext))
                # result += self.extractFile(filename)
            return result

        elif isinstance(node, yaml.MappingNode):
            result = {}
            for k,v in self.construct_mapping(node).iteritems():
                result[k] = self.extractFile(v, self._root_current)
            return result

        else:
            print "Error:: unrecognised node type in !include statement"
            raise yaml.constructor.ConstructorError

    def extractFile(self, filename, root):
        filepath = os.path.join(root, filename)
        try:
            with open(filepath, 'r') as f:
                return yaml.load(f, Loader)
        except IOError:
            pass


def replace_not_allowed_char(s):
    key = str(s)
    for s, w in {
            '/': 'SLASH',
            '*': 'STAR',
            '-': 'HYPH',
            '?': 'QUEST',
            '.': 'DOT',
            '_': 'USCORE',
    }.iteritems():
        key = key.replace(s, w)

    return int(key) if key.isdigit() else key


def gen_dict_extract(cfg, envs):
    if hasattr(cfg, 'iteritems'):
        for k, v in cfg.iteritems():
            # for final values
            if isinstance(v, (str, int, list)) and (k != 'include' and k != 'includeAfter'):
                yield {k: v}
            # for empty dict in common.yml
            if isinstance(v, dict) and len(v) == 0:
                yield {k: v}
            # for recursively descending in env/region role dict
            if k in envs and isinstance(v, (dict, list)):
                try:
                    # after descending in env main key (not the one nested under region) delete key
                    # this way when envs include both (env and region) i do not process it again
                    if k in RP_base.keys():
                        del cfg[k]
                except:
                    pass
                for result in gen_dict_extract(v, envs):
                    yield result
            # for recursively descending in dict not in RP_base_keys (env/region/envrole/stacktype) 
            # (final key is the concatenation of traversed dict keys)
            if k not in RP_base_keys and isinstance(v, (dict)):
                for j, w in v.iteritems():
                    for result in gen_dict_extract({k + j: w}, envs):
                        yield result
    if isinstance(cfg, list):
        for n in cfg:
            for result in gen_dict_extract(n, envs):
                yield result


def my_merge_dict(basedict, workdict):
    if isinstance(workdict, (str, list)):
        return workdict
    if len(basedict) > 0:
        sumdict = dict(basedict.items() + workdict.items())
    else:
        return dict(workdict.items())

    for k in sumdict.iterkeys():
        if k in workdict and isinstance(workdict[k], dict) and k in basedict and isinstance(basedict[k], Mapping):
            my_merge_dict(basedict[k], workdict[k])
        elif k in workdict:
            basedict[k] = workdict[k]
        #elif isinstance(basedict[k], dict) and 'PathPattern' in basedict[k] and basedict[k]['PathPattern'] == '':
        #    del basedict[k]
        else:
            pass

    return basedict


def get_RP_for_envs(value):
    RP = OrderedDict()

    if hasattr(value, 'iteritems'):
        for d, v in value.iteritems():
            RP[d] = get_RP_for_envs(v)
    elif isinstance(value, list) and len(value) > 0 and isinstance(value[0], OrderedDict):
        for d, v in enumerate(value):
            for i, j in v.iteritems():
                # CF Mapping allow for index only alfanumeric char, this way i can specify more elegant index in CloudFormation behaviours
                key = replace_not_allowed_char(i)
                # RP[key] already exist as a dict, try merging
                if key in RP and isinstance(RP[key], dict):
                    RP[key] = my_merge_dict(RP[key], get_RP_for_envs(j))
                else:
                    RP[key] = get_RP_for_envs(j)
    elif isinstance(value, list):
        RP = list(value)
    else:
        RP = value

    return RP


def read_yaml(file_type, brand, base_dir):
    cfg_file = os.path.join(base_dir, brand, file_type + '.yml')

    yaml.Loader.add_constructor(u'tag:yaml.org,2002:map', construct_ordereddict)
    try:
        with open(cfg_file, 'r') as ymlfile:
            cfg = yaml.load(ymlfile, Loader=Loader)

            return cfg
    except IOError:
        return {}


def parse_cfg(cfg, envs=[]):
    odict = OrderedDict([])

    for value in gen_dict_extract(cfg, envs):
        for k, v in value.iteritems():
            if k in odict and isinstance(v, list) and len(v) > 0 and isinstance(v[0], dict):
                odict[k] = odict[k] + v
            else:
                odict[k] = v

    return odict


def merge_cfg(cfgs, cfg_key, list_base=None):
    RP_list = copy.deepcopy(list_base) if list_base else []

    for cfg, v in cfgs.iteritems():
        for c in v:
            keys = ['include', 'includeAfter'] + cfg_key[cfg]
            parsed_cfg = parse_cfg(c, keys)
            RP_list.append(parsed_cfg)

    return RP_list


def get_RP(cfgs):
    RP = copy.deepcopy(RP_base)

    cfg_key_cmm = {
        'common': ['global'],
        'type': [stacktype, 'global'],
        'role': [envrole, 'global'],
    }

    RP_list = []

    cfg_merge_cmm = merge_cfg(cfgs, cfg_key_cmm)

    RP['cmm']['cmm'] = get_RP_for_envs(cfg_merge_cmm)

    for env, rvalue in RP.iteritems():
        if env == 'cmm':
            continue

        cfg_key_env = {
            'common': [env],
            'type': [env],
            'role': [env],
        }

        cfg_merge_env = merge_cfg(cfgs, cfg_key_env)

        for region in rvalue.iterkeys():

            cfg_key_region = {
                'common': [region],
                'type': [region],
                'role': [region],
            }

            cfg_key_env_region = {
                'common': [env, region],
                'type': [env, region],
                'role': [env, region],
            }

            cfg_merge_region = merge_cfg(cfgs, cfg_key_region, cfg_merge_env)
            cfg_merge_env_region = merge_cfg(cfgs, cfg_key_env_region, cfg_merge_region)

            RP[env][region] = get_RP_for_envs(cfg_merge_env_region)

    return(RP)


def get_RP_base_keys():
    RP_base_keys = ['global']
    for n, v in RP_base.iteritems():
        RP_base_keys.append(n)
        for m, w in v.iteritems():
            RP_base_keys.append(m)

    return RP_base_keys + [stacktype, envrole]


def get_stack_type(cfgs):
    for c in cfgs:
        try:
            if isinstance(c[envrole], dict) and 'StackType' in c[envrole]:
                return c[envrole]['StackType']
        except KeyError:
            pass


global envrole
global stacktype
global RP_base
global RP_base_keys
global brand
global CFG_FILE_INT
global CFG_FILE_EXT

CFG_FILE_INT = '%s/cfg' % os.path.dirname(os.path.realpath(__file__))
CFG_FILE_INT = os.path.normpath(CFG_FILE_INT)

CFG_FILE_EXT = '%s/cfg' % os.getcwd()
CFG_FILE_EXT = os.path.normpath(CFG_FILE_EXT)

envrole = cfg.envrole
brand = cfg.brand
RP_base = cfg.RP_base

cfg_role = [
    read_yaml(envrole, 'BASE', CFG_FILE_INT),
    read_yaml(envrole, 'BASE', CFG_FILE_EXT),
    read_yaml(envrole, brand, CFG_FILE_EXT),
]

stacktype = get_stack_type(cfg_role)

if not stacktype:
    logging.error('StackType key not found for Role %s in paths: %s, %s', envrole, CFG_FILE_EXT, CFG_FILE_INT)
    exit(1)

cfgs = OrderedDict([
    ('common', [
        read_yaml('common', 'BASE', CFG_FILE_INT),
        read_yaml('common', 'BASE', CFG_FILE_EXT),
        read_yaml('common', brand, CFG_FILE_EXT),
    ]),
    ('type', [
        read_yaml(stacktype, 'BASE', CFG_FILE_INT),
        read_yaml(stacktype, 'BASE', CFG_FILE_EXT),
        read_yaml(stacktype, brand, CFG_FILE_EXT),
    ]),
    ('role', cfg_role),
])

RP_base_keys = get_RP_base_keys()

RP = get_RP(cfgs)

try:
    stacktype = RP['cmm']['cmm']['StackType']
except KeyError:
    exit(1)

#    print(RP['dev']['eu-west-1']['CloudFrontCacheBehaviors'][2]['QueryStringCacheKeys'])
if cfg.debug:
    # TO DEBUG - NICELY PRINT NESTED ORDEREDDICT
    show_odict(RP)
    #pprint(RP)

cfg.RP = RP
cfg.RP_cmm = RP['cmm']['cmm']
cfg.stacktype = stacktype



import troposphere.elasticache as cch

from shared import *


class CCHCacheCluster(cch.CacheCluster):
    def setup(self):
        auto_get_props(self, RP_cmm, mapname='')


class CCHCacheClusterPublic(CCHCacheCluster):
    def setup(self):
        super(CCHCacheClusterPublic, self).setup()
        self.CacheSubnetGroupName = get_exported_value('CacheSubnetGroupPublic')
        self.CacheSecurityGroupNames = [Ref('SecurityGroupCCH')]


class CCHCacheClusterPrivate(CCHCacheCluster):
    def setup(self):
        super(CCHCacheClusterPrivate, self).setup()
        self.CacheSubnetGroupName = get_exported_value('CacheSubnetGroupPrivate')
        self.VpcSecurityGroupIds = [Ref('SecurityGroupCCH')]


class CCHCacheSubnetGroupPrivate(cch.SubnetGroup):
    def setup(self):
        self.Description = Sub('${EnvShort}-Private')
        self.SubnetIds=Split(',', get_exported_value('SubnetsPrivate'))


class CCHCacheSubnetGroupPublic(cch.SubnetGroup):
    def setup(self):
        self.Description = Sub('${EnvShort}-Public')
        self.SubnetIds=Split(',', get_exported_value('SubnetsPublic'))

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class CCH_Cache(object):
    def __init__(self, key):
        # Resources
        if RP_cmm['CCHScheme'] == 'External':
            R_Cache = CCHCacheClusterPublic('CacheCluster')
        if RP_cmm['CCHScheme'] == 'Internal':
            R_Cache = CCHCacheClusterPrivate('CacheCluster')

        R_Cache.setup()
        R_Cache.Condition = 'CacheEnabled'

        R53_RecordSetCCH()

        cfg.Resources.extend([
            R_Cache,
        ])


class CCH_SubnetGroups(object):
    def __init__(self, key):
        # Resources
        R_Private = CCHCacheSubnetGroupPrivate('CacheSubnetGroupPrivate')
        R_Private.setup()

        R_Public = CCHCacheSubnetGroupPublic('CacheSubnetGroupPublic')
        R_Public.setup()

        cfg.Resources.extend([
            R_Private,
            R_Public,
        ])

        # Outputs
        O_Private = Output('CacheSubnetGroupPrivate')
        O_Private.Value = Ref('CacheSubnetGroupPrivate')
        O_Private.Export = Export('CacheSubnetGroupPrivate')

        O_Public = Output('CacheSubnetGroupPublic')
        O_Public.Value = Ref('CacheSubnetGroupPublic')
        O_Public.Export = Export('CacheSubnetGroupPublic')

        cfg.Outputs.extend([
            O_Private,
            O_Public,
        ])

# Need to stay as last lines
import_modules(globals())



#!/usr/bin/python
import sys
import os

parent_dir_name = os.getcwd()
sys.path.append(parent_dir_name + '/lib')

import awsibox.cfg as cfg
from awsibox.args import *
from awsibox.RP import *
from awsibox.mapping import *
from awsibox.shared import *

stacktype = cfg.stacktype
envrole = cfg.envrole


def execute_class():
    for k, v in cfg.CFG_TO_CLASS.iteritems():
        if k in RP_cmm.keys():
            RP_value = RP_cmm[k]
            if isinstance(RP_value, str) and RP_value == 'SkipClass':
                continue
            if isinstance(v, list):
                for n in v:
                    globals()[n](key=k)
                continue
            if v + stacktype.upper() in globals():
                globals()[v + stacktype.upper()](key=k)
            elif v in globals():
                globals()[v](key=k)

    stack_add_res()


# ################################
# #### END STACK ROLE CLASSES ####
# ################################

classenvrole = envrole.replace('-', '_')  # Ex client-portal -> client_portal
cfg.classenvrole = classenvrole

execute_class()

cfg.template.add_description(envrole + ' [' + stacktype + ']')
cfg.template.add_version('2010-09-09')

print(cfg.template.to_json())
# print(cfg.template.to_yaml())



import troposphere.cloudfront as clf

from shared import *


class CFOriginAccessIdentity(clf.CloudFrontOriginAccessIdentity):
    def setup(self, comment):
        self.CloudFrontOriginAccessIdentityConfig = clf.CloudFrontOriginAccessIdentityConfig(
            Comment=comment,
        )

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################

# S - CLOUDFRONT #
class CFDefaultCacheBehavior(clf.DefaultCacheBehavior):
    def setup(self, key):
        name = self.title
        if 'AllowedMethods' in key:
            self.AllowedMethods = get_final_value(name + 'AllowedMethods')

        if 'CachedMethods' in key:
            self.CachedMethods = get_final_value(name + 'CachedMethods')

        # If not defined default to True
        if 'Compress' in key:
            self.Compress = get_final_value(name + 'Compress')
        else:
            self.Compress = True

        if 'DefaultTTL' in key:
            self.DefaultTTL = get_final_value(name + 'DefaultTTL')

        self.ForwardedValues = clf.ForwardedValues()
        # If not defined default to True
        if 'QueryString' in key:
            self.ForwardedValues.QueryString=get_final_value(name + 'QueryString')
        else:
            self.ForwardedValues.QueryString=True

        if 'CookiesForward' in key:
            self.ForwardedValues.Cookies = clf.Cookies(
                Forward=get_final_value(name + 'CookiesForward')
            )
            if 'CookiesWhitelistedNames' in key:
                self.ForwardedValues.Cookies.WhitelistedNames = get_final_value(name + 'CookiesWhitelistedNames', condition=True)
                # conditions
                do_no_override(True)
                cfg.Conditions.append({
                    name + 'CookiesWhitelistedNames': Equals(get_final_value(name + 'CookiesForward'), 'whitelist')
                })
                do_no_override(False)

        # If not defined default to 'Host'
        if 'Headers' in key:
            self.ForwardedValues.Headers = get_final_value(name + 'Headers')
        else:
            self.ForwardedValues.Headers = ['Host']

        if 'QueryStringCacheKeys' in key:
            self.ForwardedValues.QueryStringCacheKeys = get_final_value(name + 'QueryStringCacheKeys', condition=True)
            # conditions
            do_no_override(True)
            cfg.Conditions.append({
                name + 'QueryStringCacheKeys': Equals(get_final_value(name + 'QueryString'), True)
            })
            do_no_override(False)

        if 'TargetOriginId' in key:
            self.TargetOriginId = get_final_value(name + 'TargetOriginId')
        else:
            self.TargetOriginId = If(
                'CloudFrontOriginAdHoc',
                Ref('RecordSetExternal'),
                Sub('${EnvRole}${RecordSetCloudFrontSuffix}.origin.' + get_final_value('HostedZoneNameEnv'))
            )

        if 'MaxTTL' in key:
            self.MaxTTL = get_final_value(name + 'MaxTTL')

        if 'MinTTL' in key:
            self.MinTTL = get_final_value(name + 'MinTTL')

        if 'LambdaFunctionARN' in key:
            condname = name + 'LambdaFunctionARN'
            eventType = 'origin-request'
            if 'LambdaEventType' in key:
                eventType = key['LambdaEventType']
            # conditions
            do_no_override(True)
            cfg.Conditions.append({
                condname: Not(Equals(get_final_value(condname), 'None'))
            })
            do_no_override(False)
            self.LambdaFunctionAssociations = [
                If(
                    condname,
                    clf.LambdaFunctionAssociation(
                        EventType=eventType,
                        LambdaFunctionARN=get_final_value(condname)
                    ),
                    Ref('AWS::NoValue')
                )
            ]

        if 'ViewerProtocolPolicy' in key:
            self.ViewerProtocolPolicy = get_final_value(name + 'ViewerProtocolPolicy')


class CFCacheBehavior(clf.CacheBehavior, CFDefaultCacheBehavior):
    def setup(self, **kwargs):
        super(CFCacheBehavior, self).setup(**kwargs)
        name = self.title
        self.PathPattern = get_final_value(name + 'PathPattern')


class CFDistributionConfig(clf.DistributionConfig):
    def setup(self):
        self.Enabled = 'True'
        DefaultCacheBehavior = CFDefaultCacheBehavior('CloudFrontCacheBehaviors0')
        DefaultCacheBehavior.setup(key=RP_cmm['CloudFrontCacheBehaviors'][0])
        self.DefaultCacheBehavior = DefaultCacheBehavior
        self.HttpVersion = get_final_value('CloudFrontHttpVersion')
        self.Logging = If(
            'CloudFrontLogging',
            clf.Logging(
                Bucket=Sub(get_final_value('BucketLogs') + '.s3.amazonaws.com'),
                Prefix=Sub('${EnvRole}.${AWS::StackName}/'),
            ),
            Ref('AWS::NoValue')
        )
        self.PriceClass = 'PriceClass_100'
        self.ViewerCertificate = clf.ViewerCertificate(
            AcmCertificateArn=If(
                'CloudFrontAcmCertificate',
                get_final_value('CloudFrontAcmCertificateArn'),
                Ref('AWS::NoValue')
            ),
            CloudFrontDefaultCertificate=If(
                'CloudFrontAcmCertificate',
                Ref('AWS::NoValue'),
                'True'
            ),
            SslSupportMethod=If(
                'CloudFrontAcmCertificate',
                'sni-only',
                Ref('AWS::NoValue')
            )
        )
        self.WebACLId = get_final_value('CloudFrontWebACLId', condition=True)


class CFCustomOrigin(clf.CustomOriginConfig):
    def setup(self, key):
        name = self.title
        if 'HTTPSPort' in key:
            self.HTTPSPort = get_final_value(name + 'HTTPSPort')
        if 'HTTPSort' in key:
            self.HTTPPort = get_final_value(name + 'HTTPPort')
        self.OriginProtocolPolicy = get_final_value(name + 'ProtocolPolicy')
        if 'SSLProtocols' in key:
            self.OriginSSLProtocols = get_final_value(name + 'SSLProtocols')
        if 'ReadTimeout' in key:
            self.OriginReadTimeout = get_final_value(name + 'ReadTimeout')
        if 'KeepaliveTimeout' in key:
            self.OriginKeepaliveTimeout = get_final_value(name + 'KeepaliveTimeout')


class CFOriginCustomHeader(clf.OriginCustomHeader):
    def setup(self):
        name = self.title
        self.HeaderName = get_final_value(name + 'Name')
        self.HeaderValue = get_final_value(name + 'Value')


class CFOriginCLF(clf.Origin):
    def setup(self, index):
        name = self.title
        key = RP_cmm['CloudFrontOrigins'][index]

        CustomHeaders = []
        if 'Headers' in key:
            for n in key['Headers']:
                headername = name + 'Headers' + str(n)
                CustomHeader = CFOriginCustomHeader(headername)
                CustomHeader.setup()

                CustomHeaders.append(CustomHeader)

        if key['Type'] == 'custom':
            CustomOrigin = CFCustomOrigin(name)
            CustomOrigin.setup(key=key)
            self.CustomOriginConfig = CustomOrigin
        else:
            self.S3OriginConfig = clf.S3OriginConfig()
            if 'OriginAccessIdentity' in key:
                self.S3OriginConfig.OriginAccessIdentity = get_sub_mapex('origin-access-identity/cloudfront/${1M}', name + 'OriginAccessIdentity')

        self.DomainName = get_final_value(name + 'DomainName')
        self.OriginPath = get_final_value(name + 'Path')
        self.Id = get_final_value(name + 'Id')
        self.OriginCustomHeaders = CustomHeaders


class CFOriginFixed(clf.Origin):
    def setup(self):
        cloudfrontorigincustomheaders = []

        self.DomainName = If(
            'CloudFrontOriginAdHoc',
            Ref('RecordSetExternal'),
            Sub('${EnvRole}${RecordSetCloudFrontSuffix}.origin.' + get_final_value('HostedZoneNameEnv'))
        )
        self.Id = self.DomainName
        self.OriginPath = get_final_value('CloudFrontOriginPath')

        if 'CloudFrontOriginCustomHeaders' in RP_cmm:
            for n in RP_cmm['CloudFrontOriginCustomHeaders']:
                name = 'CloudFrontOriginCustomHeaders' + str(n)
                cloudfrontorigincustomheaders.append(
                    If(
                        name,
                        clf.OriginCustomHeader(
                            HeaderName=get_final_value(name + 'HeaderName'),
                            HeaderValue=get_final_value(name + 'HeaderValue')
                        ),
                        Ref("AWS::NoValue")
                    )
                )
                # conditions
                do_no_override(True)
                cfg.Conditions.append({
                    name: Not(Equals(get_final_value(name + 'HeaderName'), ''))
                })
                do_no_override(False)

        self.OriginCustomHeaders = cloudfrontorigincustomheaders
        self.CustomOriginConfig = clf.CustomOriginConfig(
            HTTPSPort=If(
                'CloudFrontOriginProtocolHTTPS',
                get_final_value('ListenerLoadBalancerHttpsPort'),
                Ref('AWS::NoValue')
            ),
            HTTPPort=If(
                'CloudFrontOriginProtocolHTTP',
                get_final_value('ListenerLoadBalancerHttpPort'),
                Ref('AWS::NoValue')
            ),
            OriginProtocolPolicy=get_final_value('CloudFrontOriginProtocolPolicy')
        )


class CFCustomErrorResponse(clf.CustomErrorResponse):
    def setup(self):
        name = self.title  # Ex. CustomErrorResponses1
        label = name.rstrip('0987654321')
        index = int(name.lstrip(label))

        self.ErrorCode = get_final_value(name + 'ErrorCode')
        self.ErrorCachingMinTTL = get_final_value(name + 'ErrorCachingMinTTL')
        if 'ResponsePagePath' in RP_cmm[label][index]:
            self.ResponsePagePath = get_final_value(name + 'ResponsePagePath')
        if 'ResponseCode' in RP_cmm[label][index]:
            self.ResponseCode = get_final_value(name + 'ResponseCode')
# E - CLOUDFRONT #


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


# #################################
# ### START STACK INFRA CLASSES ###
# #################################

# S - CLOUDFRONT #
class CF_CloudFront(object):
    def __init__(self):
        # Parameters
        P_Logging = Parameter('CloudFrontLogging')
        P_Logging.Description = 'CloudFront Logging - None to disable - empty for default based on env/role'

        cfg.Parameters.extend([
            P_Logging,
        ])

        # Conditions
        do_no_override(True)
        C_Logging = {'CloudFrontLogging': Or(
            And(
                Condition('CloudFrontLoggingOverride'),
                Not(Equals(Ref('CloudFrontLogging'), 'None'))
            ),
            And(
                Not(Condition('CloudFrontLoggingOverride')),
                Not(Equals(get_final_value('CloudFrontLogging'), 'None'))
            )
        )}

        C_AcmCertificate = {'CloudFrontAcmCertificate': Not(
            Equals(get_final_value('CloudFrontAcmCertificate'), 'None')
        )}

        cfg.Conditions.extend([
            C_Logging,
            C_AcmCertificate,
        ])
        do_no_override(False)

        # Resources
        CloudFrontDistribution = clf.Distribution('CloudFrontDistribution')

        DistributionConfig = CFDistributionConfig()
        DistributionConfig.setup()

        cachebehaviors = []
        sortedcachebehaviors = sorted(
            RP_cmm['CloudFrontCacheBehaviors'].iteritems(),
            key=lambda (x, y): (
                (y['Order']*1000) - len(y['PathPattern']) if 'PathPattern' in y else 0
            ) if 'Order' in y else x
        )
        itercachebehaviors = iter(sortedcachebehaviors)
        itercachebehaviors.next()
        for n, v in itercachebehaviors:
            if not v['PathPattern']:
                continue
            name = 'CloudFrontCacheBehaviors' + str(n)

            cachebehavior = CFCacheBehavior(name)
            cachebehavior.setup(key=v)

            # Create and Use Condition only if PathPattern Value differ between envs
            if name + 'PathPattern' not in mappedvalue:
                # conditions
                do_no_override(True)
                c_CacheBehavior = {name: Not(
                    Equals(get_final_value(name + 'PathPattern'), 'None')
                )}

                cfg.Conditions.append(c_CacheBehavior)
                do_no_override(False)

                cachebehaviors.append(
                    If(
                        name,
                        cachebehavior,
                        Ref("AWS::NoValue")
                    )
                )
            else:
                cachebehaviors.append(cachebehavior)

        DistributionConfig.CacheBehaviors = cachebehaviors

        cloudfrontaliasextra = []
        for n in RP_cmm['CloudFrontAliasExtra']:
            name = 'CloudFrontAliasExtra' + str(n)

            # parameters
            p_Alias = Parameter(name)
            p_Alias.Description = 'Alias extra - empty for default based on env/role'

            cfg.Parameters.append(p_Alias)
            
            cloudfrontaliasextra.append(get_final_value(name, condition=True))

            # conditions
            do_no_override(True)
            c_Alias = {name: Or(
                And(
                    Condition(name + 'Override'),
                    Not(Equals(Ref(name), 'None'))
                ),
                And(
                    Not(Condition(name + 'Override')),
                    Not(Equals(get_final_value(name), 'None'))
                )
            )}

            cfg.Conditions.append(c_Alias)
            do_no_override(False)

        CloudFrontDistribution.DistributionConfig = DistributionConfig
        CloudFrontDistribution.DistributionConfig.Aliases = cloudfrontaliasextra

        self.CloudFrontDistribution = CloudFrontDistribution


class CF_CloudFrontEC2(CF_CloudFront):
    def __init__(self, key):
        super(CF_CloudFrontEC2, self).__init__()

        # Parameters
        P_CloudFront = Parameter('CloudFront')
        P_CloudFront.Description = 'Create CloudFront Distribution - empty for default based on env/role'
        P_CloudFront.AllowedValues = ['', 'None', 'True']

        P_RecordSetCloudFrontSuffix = Parameter('RecordSetCloudFrontSuffix')
        P_RecordSetCloudFrontSuffix.Description = 'RecordSetCloudFront DNS Name Suffix - empty to disable'

        cfg.Parameters.extend([
            P_CloudFront,
            P_RecordSetCloudFrontSuffix,
        ])

        # Conditions
        do_no_override(True)
        C_AliasZone = {'CloudFrontAliasZone': Not(
            Equals(get_final_value('CloudFrontAliasZone'), 'None')
        )}

        C_Distribution = {'CloudFrontDistribution': Or(
            And(
                Condition('CloudFrontOverride'),
                Not(Equals(Ref('CloudFront'), 'None'))
            ),
            And(
                Not(Condition('CloudFrontOverride')),
                Not(Equals(get_final_value('CloudFront'), 'None'))
            )
        )}

        C_OriginAdHoc = {'CloudFrontOriginAdHoc': Equals(
            get_final_value('CloudFrontOriginAdHoc'), True
        )}

        C_OriginProtocolHTTP = {'CloudFrontOriginProtocolHTTP': And(
            Condition('ListenerLoadBalancerHttpPort'),
            Not(Equals(get_final_value('CloudFrontOriginProtocolPolicy'), 'https-only'))
        )}

        C_OriginProtocolHTTPS = {'CloudFrontOriginProtocolHTTPS': And(
            Condition('ListenerLoadBalancerHttpsPort'),
            Not(Equals(get_final_value('CloudFrontOriginProtocolPolicy'), 'http-only'))
        )}

        cfg.Conditions.extend([
            C_AliasZone,
            C_Distribution,
            C_OriginAdHoc,
            C_OriginProtocolHTTP,
            C_OriginProtocolHTTPS,
        ])
        do_no_override(False)

        # Resources
        self.CloudFrontDistribution.DistributionConfig.Aliases[0:0] = [
            If(
                'RecordSetCloudFront',
                Sub('${EnvRole}${RecordSetCloudFrontSuffix}.cdn.' + get_final_value('HostedZoneNameEnv')),
                Ref('RecordSetExternal')
            ),
            Sub('${EnvRole}${RecordSetCloudFrontSuffix}.' + get_final_value('HostedZoneNameEnv')),
            If(
                'CloudFrontAliasZone',
                Sub(get_final_value('CloudFrontAliasZone') + '.' + get_final_value('HostedZoneNameEnv')),
                Ref('AWS::NoValue')
            )
        ]

        self.CloudFrontDistribution.Condition = 'CloudFrontDistribution'

        Origin = CFOriginFixed()
        Origin.setup()
        self.CloudFrontDistribution.DistributionConfig.Origins = [Origin]
        self.CloudFrontDistribution.DistributionConfig.Comment = Sub('${AWS::StackName}-${EnvRole}')

        R_CloudFrontDistribution = self.CloudFrontDistribution

        cfg.Resources.extend([
            R_CloudFrontDistribution,
        ])

        R53_RecordSetCloudFront()

        # Outputs
        O_CloudFront = Output('CloudFront')
        O_CloudFront.Value = get_final_value('CloudFront')

        cfg.Outputs.extend([
            O_CloudFront,
        ])


class CF_CloudFrontECS(CF_CloudFrontEC2):
    def __init__(self, key):
        super(CF_CloudFrontECS, self).__init__(key)


class CF_CloudFrontCLF(CF_CloudFront):
    def __init__(self, key):
        super(CF_CloudFrontCLF, self).__init__()

        Origins = []
        for n, v in RP_cmm['CloudFrontOrigins'].iteritems():
            name = 'CloudFrontOrigins' + str(n)
            
            # parameters
            p_Origin = Parameter('CloudFrontOrigins' + n + 'DomainName')
            p_Origin.Description = 'Origin - empty for default based on env/role'

            p_OriginHTTPSPort = Parameter('CloudFrontOrigins' + n + 'HTTPSPort')
            p_OriginHTTPSPort.Description = 'Origin - empty for default based on env/role'

            cfg.Parameters.extend([
                p_Origin,
                p_OriginHTTPSPort,
            ])

            Origin = CFOriginCLF(name)
            Origin.setup(index=n)
            Origins.append(Origin)

        CustomErrorResponses = []
        if hasattr(RP_cmm['CustomErrorResponses'], 'iteritems'):
            for n in RP_cmm['CustomErrorResponses']:
                name = 'CustomErrorResponses' + str(n)
                CustomErrorResponse = CFCustomErrorResponse(name)
                CustomErrorResponse.setup()
                CustomErrorResponses.append(CustomErrorResponse)

        self.CloudFrontDistribution.DistributionConfig.Origins = Origins
        self.CloudFrontDistribution.DistributionConfig.CustomErrorResponses = CustomErrorResponses
        self.CloudFrontDistribution.DistributionConfig.Comment = get_final_value('CloudFrontComment')

        R_CloudFrontDistribution = self.CloudFrontDistribution

        cfg.Resources.extend([
            R_CloudFrontDistribution,
        ])
# E - CLOUDFRONT #

# Need to stay as last lines
import_modules(globals())



import troposphere.route53 as r53

from shared import *


# S - ROUTE53 #
class R53RecordSet(r53.RecordSetType):
    def setup(self):
        pass


class R53RecordSetZoneExternal(R53RecordSet):
    def setup(self):
        super(R53RecordSetZoneExternal, self).setup()
        self.HostedZoneId = get_exported_value('HostedZoneIdEnv')
        self.Name = Sub('${AWS::StackName}.${EnvRole}.' + get_final_value('HostedZoneNameRegionEnv'))  # Ex. prt-a-d.client-portal.eu-west-1.dev..


class R53RecordSetZoneInternal(R53RecordSet):
    def setup(self):
        super(R53RecordSetZoneInternal, self).setup()
        self.HostedZoneId = get_exported_value('HostedZoneIdPrivate')
        self.Name = Sub('${AWS::StackName}.${EnvRole}.' + get_final_value('HostedZoneNamePrivate'))  # Ex. prt-a-d.client-portal.internal..


class R53RecordSetCloudFront(R53RecordSetZoneExternal):
    def setup(self):
        super(R53RecordSetCloudFront, self).setup()
        self.Condition = 'RecordSetCloudFront'
        self.AliasTarget = r53.AliasTarget(
            DNSName=GetAtt('CloudFrontDistribution', 'DomainName'),
            HostedZoneId=get_final_value('HostedZoneIdCF')
        )
        self.Name = Sub('${EnvRole}${RecordSetCloudFrontSuffix}.cdn.' + get_final_value('HostedZoneNameEnv'))  # Ex. client-portal.cdn.dev..
        self.Type = 'A'


class R53RecordSetLoadBalancer(R53RecordSet):
    def setup(self):
        super(R53RecordSetLoadBalancer, self).setup()
        self.AliasTarget = r53.AliasTarget(
            HostedZoneId=get_final_value('HostedZoneIdLB')
        )
        self.Type = 'A'


class R53RecordSetEC2LoadBalancerExternal(R53RecordSetLoadBalancer, R53RecordSetZoneExternal):
    pass


class R53RecordSetEC2LoadBalancerInternal(R53RecordSetLoadBalancer, R53RecordSetZoneInternal):
    pass


class R53RecordSetECSLoadBalancer(R53RecordSetLoadBalancer):
    def setup(self, scheme):
        super(R53RecordSetECSLoadBalancer, self).setup()
        #scheme = RP_cmm['LoadBalancerApplication']
        self.AliasTarget.DNSName = get_sub_mapex(
            'dualstack.${1E}',
            'LoadBalancerApplication' + scheme + 'DNS',
            'LoadBalancerApplicationStack'
        ) 


class R53RecordSetECSLoadBalancerApplicationExternal(R53RecordSetECSLoadBalancer, R53RecordSetZoneExternal):
    pass


class R53RecordSetECSLoadBalancerApplicationInternal(R53RecordSetECSLoadBalancer, R53RecordSetZoneInternal):
    pass


class R53RecordSetEFS(R53RecordSet):
    def setup(self, efsname):
        super(R53RecordSetEFS, self).setup()
        condname = 'EFSFileSystem' + efsname  # Ex. EFSFileSystemWordPress
        self.Condition = condname
        self.HostedZoneId = Ref('HostedZonePrivate')
        self.Name = Sub('efs-' + efsname + '.' + get_final_value('HostedZoneNamePrivate'))
        self.ResourceRecords = [
            Sub('${' + condname + '}.efs.${AWS::Region}.amazonaws.com')
        ]
        self.Type = 'CNAME'
        self.TTL = '300'


class R53RecordSetRDS(R53RecordSet):
    def setup(self):
        super(R53RecordSetRDS, self).setup()
        self.Type = 'CNAME'
        self.ResourceRecords = [GetAtt('DBInstance', 'Endpoint.Address')]
        self.TTL = '300'


class R53RecordSetRDSExternal(R53RecordSetRDS, R53RecordSetZoneExternal):
    pass


class R53RecordSetRDSInternal(R53RecordSetRDS, R53RecordSetZoneInternal):
    pass


class R53RecordSetCCH(R53RecordSet):
    def setup(self):
        super(R53RecordSetCCH, self).setup()
        self.Condition = 'CacheEnabled'
        self.Type = 'CNAME'
        if RP_cmm['Engine'] == 'memcached':
            self.ResourceRecords = [GetAtt('CacheCluster', 'ConfigurationEndpoint.Address')]
        if RP_cmm['Engine'] == 'redis':
            self.ResourceRecords = [GetAtt('CacheCluster', 'RedisEndpoint.Address')]
        self.TTL = '300'


class R53RecordSetCCHExternal(R53RecordSetCCH, R53RecordSetZoneExternal):
    pass


class R53RecordSetCCHInternal(R53RecordSetCCH, R53RecordSetZoneInternal):
    pass


class R53RecordSetNSServiceDiscovery(R53RecordSet):
    def setup(self):
        self.HostedZoneId = Ref('HostedZoneEnv')
        self.Name = Sub('find.' + RP_cmm['HostedZoneNameEnv'])
        self.ResourceRecords = GetAtt('PublicDnsNamespace', 'NameServers')
        self.Type = 'NS'
        self.TTL = '300'


class R53HostedZonePrivate(r53.HostedZone):
    def setup(self):
        self.HostedZoneConfig = r53.HostedZoneConfiguration(
            Comment=Sub('${EnvShort} private zone ${AWS::Region}')
        )
        self.Name = Sub(get_final_value('HostedZoneNamePrivate'))
        self.VPCs = [
            r53.HostedZoneVPCs(
                VPCId=get_exported_value('VpcId'),
                VPCRegion=Ref('AWS::Region')
            )
        ]


class R53HostedZoneEnv(r53.HostedZone):
    def setup(self):
        self.Condition = self.title  # Ex. HostedZoneEnv
        self.HostedZoneConfig = r53.HostedZoneConfiguration(
            Comment=Sub('${EnvShort} public zone')
        )
        self.Name = Sub(get_final_value('HostedZoneNameEnv'))


class R53HostedZoneEnvExtra1(r53.HostedZone):
    def setup(self):
        self.Condition = self.title  # Ex. HostedZoneEnvExtra1
        self.HostedZoneConfig = r53.HostedZoneConfiguration(
            Comment=Sub('${EnvShort} attico public zone')
        )
        self.Name = Sub('${EnvShort}.attico.it')

# E - ROUTE53 #

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

# S - ROUTE53 #
class R53_RecordSetCloudFront(object):
    def __init__(self):
        # Conditions
        C_RecordSet = {'RecordSetCloudFront': And(
            Condition('CloudFrontDistribution'),
            Not(
                Equals(get_final_value('RecordSetCloudFront'), 'None')
            )
        )}

        cfg.Conditions.extend([
            C_RecordSet,
        ])

        # Resources
        R_RecordSet = R53RecordSetCloudFront('RecordSetCloudFront')
        R_RecordSet.setup()

        cfg.Resources.extend([
            R_RecordSet,
        ])

        # Outputs
        O_CloudFront = Output('RecordSetCloudFront')
        O_CloudFront.Condition = 'RecordSetCloudFront'
        O_CloudFront.Value = Sub('${RecordSet} --> ${CloudFrontDistribution.DomainName}', **{'RecordSet': R_RecordSet.Name})

        cfg.Outputs.extend([
            O_CloudFront
        ])


class R53_RecordSetEC2LoadBalancer(object):
    def __init__(self):
        # Resources
        if 'External' in RP_cmm['RecordSet']:
            R_External = R53RecordSetEC2LoadBalancerExternal('RecordSetExternal')
            R_External.setup()
            if 'LoadBalancerClassic' in RP_cmm:
                if 'External' in RP_cmm['LoadBalancerClassic']:
                    R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicExternal.DNSName}')
                elif 'Internal' in RP_cmm['LoadBalancerClassic']:
                    R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicInternal.DNSName}')
            if 'LoadBalancerApplication' in RP_cmm:
                if 'External' in RP_cmm['LoadBalancerApplication']:
                    R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationExternal.DNSName}')
                elif 'Internal' in RP_cmm['LoadBalancerApplication']:
                    R_External.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationInternal.DNSName}')

            cfg.Resources.append(R_External)

            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')

            cfg.Outputs.append(O_External)

        if 'Internal' in RP_cmm['RecordSet']:
            R_Internal = R53RecordSetEC2LoadBalancerInternal('RecordSetInternal')
            R_Internal.setup()
            if 'LoadBalancerClassic' in RP_cmm:
                if 'Internal' in RP_cmm['LoadBalancerClassic']:
                    R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicInternal.DNSName}')
                elif 'External' in RP_cmm['LoadBalancerClassic']:
                    R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerClassicExternal.DNSName}')
            if 'LoadBalancerApplication' in RP_cmm:
                if 'Internal' in RP_cmm['LoadBalancerApplication']:
                    R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationInternal.DNSName}')
                elif 'External' in RP_cmm['LoadBalancerApplication']:
                    R_Internal.AliasTarget.DNSName = Sub('dualstack.${LoadBalancerApplicationExternal.DNSName}')

            cfg.Resources.append(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')

            cfg.Outputs.append(O_Internal)
# E - ROUTE53 #


class R53_RecordSetECSLoadBalancer(object):
    def __init__(self):
        # Resources
        if 'External' in RP_cmm['RecordSet']:
            R_External = R53RecordSetECSLoadBalancerApplicationExternal('RecordSetExternal')

            if 'External' in RP_cmm['LoadBalancerApplication']:
                R_External.setup(scheme='External')
            else:
                R_External.setup(scheme='Internal')

            cfg.Resources.append(R_External)
            
            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')

            cfg.Outputs.append(O_External)

        if 'Internal' in RP_cmm['RecordSet']:
            R_Internal = R53RecordSetECSLoadBalancerApplicationInternal('RecordSetInternal')

            if 'Internal' in RP_cmm['LoadBalancerApplication']:
                R_Internal.setup(scheme='Internal')
            else:
                R_Internal.setup(scheme='External')

            cfg.Resources.append(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')

            cfg.Outputs.append(O_Internal)


class R53_RecordSetRDS(object):
    def __init__(self):
        # Resources
        if 'External' in RP_cmm['RecordSet']:
            R_External = R53RecordSetRDSExternal('RecordSetExternal')
            R_External.setup()
            cfg.Resources.append(R_External)

            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')

            cfg.Outputs.append(O_External)

        if 'Internal' in RP_cmm['RecordSet']:
            R_Internal = R53RecordSetRDSInternal('RecordSetInternal')
            R_Internal.setup()
            cfg.Resources.append(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')

            cfg.Outputs.append(O_Internal)


class R53_RecordSetCCH(object):
    def __init__(self):
        # Resources
        if 'External' in RP_cmm['RecordSet']:
            R_External = R53RecordSetCCHExternal('RecordSetExternal')
            R_External.setup()
            cfg.Resources.append(R_External)

            # outputs
            O_External = Output('RecordSetExternal')
            O_External.Value = Ref('RecordSetExternal')
            O_External.Condition = 'CacheEnabled'

            cfg.Outputs.append(O_External)

        if 'Internal' in RP_cmm['RecordSet']:
            R_Internal = R53RecordSetCCHInternal('RecordSetInternal')
            R_Internal.setup()
            cfg.Resources.append(R_Internal)

            # outputs
            O_Internal = Output('RecordSetInternal')
            O_Internal.Value = Ref('RecordSetInternal')
            O_Internal.Condition = 'CacheEnabled'

            cfg.Outputs.append(O_Internal)


class R53_HostedZones(object):
    def __init__(self, key):
        # Conditions
        do_no_override(True)
        C_Env = {'HostedZoneEnv': Not(
            Equals(get_final_value('HostedZoneEnv'), 'None')
        )}

        
        C_EnvExtra1 = {'HostedZoneEnvExtra1': Not(
            Equals(get_final_value('HostedZoneEnvExtra1'), 'None')
        )}
        do_no_override(False)

        cfg.Conditions.extend([
            C_Env,
            C_EnvExtra1,
        ])

        # Resources
        R_Private = R53HostedZonePrivate('HostedZonePrivate')
        R_Private.setup()

        R_Env = R53HostedZoneEnv('HostedZoneEnv')
        R_Env.setup()

        R_EnvExtra1 = R53HostedZoneEnvExtra1('HostedZoneEnvExtra1')
        R_EnvExtra1.setup()

        cfg.Resources.extend([
            R_Private,
            R_Env,
            R_EnvExtra1,
        ])

        # Outputs
        O_Private = Output('HostedZoneIdPrivate')
        O_Private.Value = Ref('HostedZonePrivate')
        O_Private.Export = Export('HostedZoneIdPrivate')

        O_PrivateName = Output('HostedZoneNamePrivate')
        O_PrivateName.Value =  Sub(get_final_value('HostedZoneNamePrivate'))

        O_Env = Output('HostedZoneIdEnv')
        O_Env.Value = If(
            'HostedZoneEnv',
            Ref('HostedZoneEnv'),
            get_final_value('HostedZoneIdEnv')
        )
        O_Env.Export = Export('HostedZoneIdEnv')

        O_EnvName = Output('HostedZoneNameEnv')
        O_EnvName.Value = Sub(get_final_value('HostedZoneNameEnv'))

        O_EnvNameRegion = Output('HostedZoneNameRegionEnv')
        O_EnvNameRegion.Value = Sub(get_final_value('HostedZoneNameRegionEnv'))

        cfg.Outputs.extend([
            O_Private,
            O_PrivateName,
            O_Env,
            O_EnvName,
            O_EnvNameRegion,
        ])

# Need to stay as last lines
import_modules(globals())



import troposphere.efs as efs

from shared import *


class EFSFileSystem(efs.FileSystem):
    def setup(self):
        self.Condition = self.title
        self.Encrypted = False
        self.PerformanceMode = 'generalPurpose'


class EFSMountTarget(efs.MountTarget):
    def setup(self, index, sgname, efsresname):
        name = self.title
        self.Condition = name
        self.FileSystemId = Ref(efsresname)
        self.SecurityGroups = [
            Ref(sgname)
        ]
        self.SubnetId = Select(str(index), Split(',', get_exported_value('SubnetsPrivate')))

# #################################
# ### START STACK INFRA CLASSES ###
# #################################

class EFS_FileStorage(object):
    def __init__(self, key):
        for n, v in RP_cmm[key].iteritems():
            resname =  key + n  # Ex. EFSFileSystemWordPress
            recordname = 'RecordSetEFS' + n  # Ex. RecordSetEFSWordPress
            sgservername = 'SecurityGroupEFSServer' + n  # Ex. SecurityGroupEFSServerWordPress
            sgclientname = 'SecurityGroupEFS' + n  # Ex. SecurityGroupEFSWordPress
            sginame = 'SecurityGroupIngressEFS' + n  # Ex. SecurityGroupIngressEFSWordPress
            for i in range(3):
                mountname = 'EFSMountTarget' + n + str(i)  # Ex. EFSMountTargetWordPress3
                # conditions
                do_no_override(True)
                c_FileZone = {mountname: And(
                    Condition(resname),
                    Equals(FindInMap('AvabilityZones', Ref('AWS::Region'), 'Zone' + str(i)), 'True')
                )}
                cfg.Conditions.append(c_FileZone)
                do_no_override(False)

                # resources
                r_Mount = EFSMountTarget(mountname)
                r_Mount.setup(index=i, sgname=sgservername, efsresname=resname)

                cfg.Resources.append(r_Mount)
            # conditions
            do_no_override(True)
            c_File = {resname: Not(
                Equals(get_final_value(resname + 'Enabled'), 'None')
            )}
            cfg.Conditions.append(c_File)
            do_no_override(False)

            # resources
            r_File = EFSFileSystem(resname)
            r_File.setup()

            r_Record = R53RecordSetEFS(recordname)
            r_Record.setup(n)

            r_SGServer = SecurityGroup(sgservername)
            r_SGServer.setup()
            r_SGServer.Condition = resname
            r_SGServer.GroupDescription = 'Rule to access EFS FileSystem ' + n

            r_SGClient = SecurityGroup(sgclientname)
            r_SGClient.setup()
            r_SGClient.Condition = resname
            r_SGClient.GroupDescription = 'Enable access to EFS FileSystem ' + n

            SGIExtra = {
                'Name': sginame,
                'Condition': resname,
                'FromPort': 2049,
                'GroupId': 'Sub(\'${' + sgservername  + '.GroupId}\')',
                'SourceSecurityGroupId': 'Ref(\'' + sgclientname + '\')',
                'ToPort': 2049
            }

            r_SGI = SecurityGroupIngress(sginame)
            auto_get_props(r_SGI, SGIExtra, rootdict=SGIExtra, mapname='', del_prefix=sginame)

            cfg.Resources.extend([
                r_File,
                r_Record,
                r_SGServer,
                r_SGClient,
                r_SGI,
            ])

            # outputs

            o_SG = Output(sgclientname)
            o_SG.Condition = resname
            o_SG.Value = GetAtt(sgclientname, 'GroupId')
            o_SG.Export = Export(sgclientname)

            cfg.Outputs.append(o_SG)

# Need to stay as last lines
import_modules(globals())
